from django.core.cache import caches
from django.template.context import make_context
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import get_template
from django.utils.html import format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import get_language_from_request

from classytags.utils import flatten_context
from djangocms_text_ckeditor.utils import OBJ_ADMIN_RE

from cmsplugin_cascade import app_settings
from cmsplugin_cascade.mixins import CascadePluginMixin

__all__ = ['register_stride', 'StrideContentRenderer']


class EmulateQuerySet:
    def __init__(self, elements):
        self.elements = elements

    def all(self):
        for id, glossary in enumerate(self.elements, 1):
            yield type(str('StrideInlineElement'), (object,), {'id': id, 'glossary': glossary})()


class StrideElementBase:
    """
    Emulate a CascadeElement to be used by the CascadeContentRenderer instead of the CMSContentRenderer.
    """
    def __init__(self, plugin, data, children_data, parent=None):
        self.plugin = plugin
        self.id = data.get('pk')
        self.glossary = data.get('glossary', {})
        self.sortinline_elements = self.inline_elements = EmulateQuerySet(data.get('inlines', []))
        self.children_data = children_data
        self.parent = parent

    @property
    def pk(self):
        return self.id

    @property
    def plugin_class(self):
        return self.plugin.__class__

    def child_plugin_instances(self):
        for plugin_type, data, children_data in self.children_data:
            plugin_class = strides_plugin_map.get(plugin_type)
            element_class = strides_element_map.get(plugin_type)
            if element_class:
                yield element_class(plugin_class(), data, children_data, parent=self)

    def get_num_children(self):
        return len(self.children_data)

    def get_complete_glossary(self):
        if not hasattr(self, '_complete_glossary_cache'):
            self._complete_glossary_cache = self.get_parent_glossary().copy()
            self._complete_glossary_cache.update(self.glossary or {})
        return self._complete_glossary_cache

    def get_parent_glossary(self):
        if self.parent:
            return self.parent.get_complete_glossary()
        return {}
        # TODO: use self.placeholder.glossary as the starting dictionary
        # template = self.placeholder.page.template if self.placeholder.page else None
        # return get_placeholder_conf('glossary', self.placeholder.slot, template=template, default={})

    @property
    def tag_type(self):
        return self.plugin_class.get_tag_type(self)

    @property
    def css_classes(self):
        css_classes = self.plugin_class.get_css_classes(self)
        return mark_safe(' '.join(c for c in css_classes if c))

    @property
    def inline_styles(self):
        inline_styles = self.plugin_class.get_inline_styles(self)
        return format_html_join(' ', '{0}: {1};', (s for s in inline_styles.items() if s[1]))

    @property
    def html_tag_attributes(self):
        attributes = self.plugin_class.get_html_tag_attributes(self)
        joined = format_html_join(' ', '{0}="{1}"', ((attr, val) for attr, val in attributes.items() if val))
        if joined:
            return mark_safe(' ' + joined)
        return ''


class TextStrideElement:
    def __init__(self, plugin, data, children_data, parent=None):
        self.plugin = plugin
        self.pk = data.get('pk')
        self.body = data.get('body')
        self.children_data = children_data
        self.parent = parent

    def tags_to_user_html(self, context, placeholder):
        content_renderer = context['cms_content_renderer']
        children_instances = {}
        for plugin_type, data, children_data in self.children_data:
            plugin_class = strides_plugin_map.get(plugin_type)
            element_class = strides_element_map.get(plugin_type)
            if element_class and 'pk' in data:
                sub_plugin = plugin_class()
                children_instances[data['pk']] = element_class(sub_plugin, data, children_data, parent=self)

        def _render_tag(m):
            plugin_id = int(m.groupdict()['pk'])
            instance = children_instances[plugin_id]
            with context.push():
                sub_context = instance.plugin.render(context, instance, placeholder)
                return content_renderer.render_plugin(instance, sub_context)

        return OBJ_ADMIN_RE.sub(_render_tag, self.body)


class StridePluginBase(CascadePluginMixin):
    """
    Whenever djangocms-cascade is used in readonly mode, all Cascade plugins are instantiated a second time
    where class CascadePluginBase is replaced against this class in order to remove its dependency to django-CMS.
    """

    def __init__(self, model=None, admin_site=None, glossary_fields=None):
        if isinstance(glossary_fields, (list, tuple)):
            self.glossary_fields = list(glossary_fields)
        elif not hasattr(self, 'glossary_fields'):
            self.glossary_fields = []

    @classmethod
    def super(cls, klass, instance):
        return super(strides_plugin_map[klass.__name__], instance)

    def render(self, context, instance, placeholder):
        context.update({
            'instance': instance,
        })
        return context

    def _get_render_template(self, context, instance, placeholder):
        if hasattr(self, 'get_render_template'):
            template = self.get_render_template(context, instance, placeholder)
        elif getattr(self, 'render_template', False):
            template = getattr(self, 'render_template', False)
        else:
            template = None

        if not template:
            raise TemplateDoesNotExist("plugin {} has no render_template".format(self.__class__))
        return template

    def in_edit_mode(self, request, placeholder):
        return False

    def get_previous_instance(self, obj):
        if obj and obj.parent:
            for pos, sibling in enumerate(obj.parent.children_data):
                if sibling[1].get('pk') == obj.pk and pos > 0:
                    prev_pt, prev_data, prev_cd = obj.parent.children_data[pos - 1]
                    return strides_plugin_map[prev_pt]()

    def get_next_instance(self, obj):
        if obj and obj.parent:
            for pos, sibling in enumerate(obj.parent.children_data):
                if sibling[1].get('pk') == obj.pk and pos < len(obj.parent.children_data):
                    next_pt, next_data, next_cd = obj.parent.children_data[pos + 1]
                    return strides_plugin_map[next_pt]()


class TextStridePlugin(StridePluginBase):
    render_template = 'cms/plugins/text.html'

    def render(self, context, instance, placeholder):
        context.update({
            'body': instance.tags_to_user_html(context, placeholder),
        })
        return context


class StrideContentRenderer:
    def __init__(self, request):
        self.request = request
        self.language = get_language_from_request(request)
        self._cached_templates = {}

    def render_cascade(self, context, tree_data):
        contents = []
        # create temporary copy of context to prevent pollution for other CMS placeholders
        context = make_context(flatten_context(context))
        for plugin_type, data, children_data in tree_data.get('plugins', []):
            plugin_class = strides_plugin_map.get(plugin_type)
            element_class = strides_element_map.get(plugin_type)
            plugin_instance = element_class(plugin_class(), data, children_data)
            # create a temporary object to store the plugins cache status
            cms_cachable_plugins = type(str('CachablePlugins'), (object,), {'value': True})
            context.push(cms_cachable_plugins=cms_cachable_plugins)
            contents.append(self.render_plugin(plugin_instance, context))
        return mark_safe(''.join(contents))

    def render_plugin(self, instance, context, placeholder=None, editable=False):
        from sekizai.helpers import get_varname as get_sekizai_context_key

        sekizai_context_key = get_sekizai_context_key()
        if app_settings.CMSPLUGIN_CASCADE['cache_strides'] and getattr(instance.plugin, 'cache', not editable):
            cache = caches['default']
            key = 'cascade_element-{}'.format(instance.pk)
            content = cache.get(key)
            if content:
                context[sekizai_context_key]['css'].extend(cache.get(key + ':css_list', []))
                context[sekizai_context_key]['js'].extend(cache.get(key + ':js_list', []))
                return content
        else:
            context['cms_cachable_plugins'].value = False
        context = instance.plugin.render(context, instance, placeholder)
        context = flatten_context(context)

        template = instance.plugin._get_render_template(context, instance, placeholder)
        template = self.get_cached_template(template)
        content = template.render(context)
        if context['cms_cachable_plugins'].value:
            cache.set(key, content)
            cache.set(key + ':css_list', context[sekizai_context_key]['css'].data)
            cache.set(key + ':js_list', context[sekizai_context_key]['js'].data)
        return content

    def user_is_on_edit_mode(self):
        return False

    def get_cached_template(self, template):
        if hasattr(template, 'render'):
            return template

        if not template in self._cached_templates:
            self._cached_templates[template] = get_template(template)
        return self._cached_templates[template]


def register_stride(name, bases, attrs, model_mixins):
    # create a fake plugin class
    plugin_bases = tuple(strides_plugin_map.get(b.__name__, b) for b in bases)
    if name == 'CascadePluginBase':
        plugin_bases += (StridePluginBase,)
        strides_plugin_map[name] = type(str('StridePluginBase'), plugin_bases, {})
    else:
        strides_plugin_map[name] = type(name, plugin_bases, attrs)

        # create a corresponding stride element class
        element_bases = model_mixins + (StrideElementBase,)
        strides_element_map[name] = type(str(name + 'Element'), element_bases, {})


strides_plugin_map = {
    'TextPlugin': TextStridePlugin,
}
strides_element_map = {
    'TextPlugin': TextStrideElement,
}



import logging
from django.utils.translation import gettext_lazy as _

from cmsplugin_cascade import app_settings
from cmsplugin_cascade.utils import (compute_aspect_ratio, get_image_size, parse_responsive_length,
   compute_aspect_ratio_with_glossary)

logger = logging.getLogger('cascade')

IMAGE_RESIZE_OPTIONS = [
    ('upscale', _("Upscale image")),
    ('crop', _("Crop image")),
    ('subject_location', _("With subject location")),
    ('high_resolution', _("Optimized for Retina")),
]

IMAGE_SHAPE_CHOICES = [
    ('img-fluid', _("Responsive")),
    ('rounded', _('Rounded')),
    ('rounded-circle', _('Circle')),
    ('img-thumbnail', _('Thumbnail')),
]


def get_image_tags(instance):
    """
    Create a context returning the tags to render an ``<img ...>`` element with
    ``sizes``, ``srcset``, a fallback ``src`` and if required inline styles.
    """
    if hasattr(instance, 'image') and hasattr(instance.image, 'exif'):
        aspect_ratio = compute_aspect_ratio(instance.image)
    elif 'image' in instance.glossary and 'width' in instance.glossary['image']:
        aspect_ratio = compute_aspect_ratio_with_glossary(instance.glossary)
    else:
        # if accessing the image file fails or fake image fails, abort here
        raise FileNotFoundError("Unable to compute aspect ratio of image")

    is_responsive = 'img-fluid' in instance.glossary.get('image_shapes', [])
    resize_options = instance.glossary.get('resize_options', {})
    crop = 'crop' in resize_options
    upscale = 'upscale' in resize_options
    if 'subject_location' in resize_options and hasattr(instance.image, 'subject_location'):
        subject_location = instance.image.subject_location
    else:
        subject_location = None
    tags = {'sizes': [], 'srcsets': {}, 'is_responsive': is_responsive, 'extra_styles': {}}
    if is_responsive:
        image_width = parse_responsive_length(instance.glossary.get('image_width_responsive') or '100%')
        assert(image_width[1]), "The given image has no valid width"
        if image_width[1] != 1.0:
            tags['extra_styles'].update({'max-width': '{:.0f}%'.format(100 * image_width[1])})
    else:
        image_width = parse_responsive_length(instance.glossary['image_width_fixed'])
        if not image_width[0]:
            image_width = (instance.image.width, image_width[1])
    try:
        image_height = parse_responsive_length(instance.glossary['image_height'])
    except KeyError:
        image_height = (None, None)
    if is_responsive:
        column_bounds_min = instance.glossary['column_bounds']['min']
        if 'high_resolution' in resize_options:
            column_bounds_max = 2 * instance.glossary['column_bounds']['max']
        else:
            column_bounds_max = instance.glossary['column_bounds']['max']
        num_steps = min(int((column_bounds_max - column_bounds_min) / app_settings.RESPONSIVE_IMAGE_STEP_SIZE),
                        app_settings.RESPONSIVE_IMAGE_MAX_STEPS)
        step_width, max_width = (column_bounds_max - column_bounds_min) / num_steps, 0
        for step in range(0, num_steps + 1):
            width = round(column_bounds_min + step_width * step)
            max_width = max(max_width, width)
            size = get_image_size(width, image_height, aspect_ratio)
            key = '{0}w'.format(*size)
            tags['srcsets'][key] = {'size': size, 'crop': crop, 'upscale': upscale,
                                    'subject_location': subject_location}
        tags['sizes'] = instance.glossary['media_queries'].values()
        # use an existing image as fallback for the <img ...> element
        if not max_width > 0:
            logger.warning('image tags: image max width is zero')
        size = (int(round(max_width)), int(round(max_width * aspect_ratio)))
    else:
        size = get_image_size(image_width[0], image_height, aspect_ratio)
        if 'high_resolution' in resize_options:
            tags['srcsets']['1x'] = {'size': size, 'crop': crop, 'upscale': upscale,
                                     'subject_location': subject_location}
            tags['srcsets']['2x'] = dict(tags['srcsets']['1x'], size=(size[0] * 2, size[1] * 2))
    tags['src'] = {'size': size, 'crop': crop, 'upscale': upscale, 'subject_location': subject_location}
    return tags


def get_picture_elements(instance):
    """
    Create a context, used to render a <picture> together with all its ``<source>`` elements:
    It returns a list of HTML elements, each containing the information to render a ``<source>``
    element.
    The purpose of this HTML entity is to display images with art directions. For normal images use
    the ``<img>`` element.
    """

    if hasattr(instance, 'image') and hasattr(instance.image, 'exif'):
        aspect_ratio = compute_aspect_ratio(instance.image)
    elif 'image' in instance.glossary and 'width' in instance.glossary['image']:
        aspect_ratio = compute_aspect_ratio_with_glossary(instance.glossary)
    else:
        # if accessing the image file fails or fake image fails, abort here
        logger.warning("Unable to compute aspect ratio of image '{}'".format(instance.image))
        return

    # container_max_heights = instance.glossary.get('container_max_heights', {})
    resize_options = instance.glossary.get('resize_options', {})
    crop = 'crop' in resize_options
    upscale = 'upscale' in resize_options
    if 'subject_location' in resize_options and hasattr(instance.image, 'subject_location'):
        subject_location = instance.image.subject_location
    else:
        subject_location = None
    max_width = 0
    max_zoom = 0
    elements = []
    for bp, media_query in instance.glossary['media_queries'].items():
        width, media = media_query['width'], media_query['media']
        max_width = max(max_width, width)
        size = None
        try:
            image_height = parse_responsive_length(instance.glossary['responsive_heights'][bp])
        except KeyError:
            image_height = (None, None)
        if image_height[0]:  # height was given in px
            size = (int(width), image_height[0])
        elif image_height[1]:  # height was given in %
            size = (int(width), int(round(width * aspect_ratio * image_height[1])))
        try:
            zoom = int(
                instance.glossary['responsive_zoom'][bp].strip().rstrip('%')
            )
        except (AttributeError, KeyError, ValueError):
            zoom = 0
        max_zoom = max(max_zoom, zoom)
        if size is None:
            # as fallback, adopt height to current width
            size = (int(width), int(round(width * aspect_ratio)))
        elem = {'tag': 'source', 'size': size, 'zoom': zoom, 'crop': crop,
                'upscale': upscale, 'subject_location': subject_location, 'media': media}
        if 'high_resolution' in resize_options:
            elem['size2'] = (size[0] * 2, size[1] * 2)
        elements.append(elem)

    # add a fallback image for old browsers which can't handle the <source> tags inside a <picture> element
    if image_height[1]:
        size = (int(max_width), int(round(max_width * aspect_ratio * image_height[1])))
    else:
        size = (int(max_width), int(round(max_width * aspect_ratio)))
    elements.append({'tag': 'img', 'size': size, 'zoom': max_zoom, 'crop': crop,
                     'upscale': upscale, 'subject_location': subject_location})
    return elements



CASCADE_PLUGINS = ['custom_snippet', 'heading', 'horizontal_rule', 'simple_wrapper', 'text_image']

def set_defaults(config):
    from cmsplugin_cascade.extra_fields.config import PluginExtraFieldsConfig

    config.setdefault('plugins_with_extra_fields', {})
    plugins_with_extra_fields = config['plugins_with_extra_fields']
    plugins_with_extra_fields.setdefault('HorizontalRulePlugin', PluginExtraFieldsConfig(
        inline_styles={
            'extra_fields:Border': ['border-top'],
            'extra_fields:Border Radius': ['border-radius'],
            'extra_units:Border Radius': 'px,rem',
        },
        allow_override=False,
    ))



from django.utils.translation import gettext_lazy as _
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.plugin_base import TransparentContainer, TransparentWrapper
from cmsplugin_cascade.bootstrap4.plugin_base import BootstrapPluginBase


class CardChildBase(BootstrapPluginBase):
    require_parent = True
    parent_classes = ['BootstrapCardPlugin']
    allow_children = True
    render_template = 'cascade/generic/wrapper.html'
    child_classes = ['BootstrapCardHeaderPlugin', 'BootstrapCardBodyPlugin', 'BootstrapCardFooterPlugin']


class BootstrapCardHeaderPlugin(TransparentContainer, CardChildBase):
    name = _("Card Header")
    default_css_class = 'card-header'

plugin_pool.register_plugin(BootstrapCardHeaderPlugin)


class BootstrapCardBodyPlugin(TransparentContainer, CardChildBase):
    name = _("Card Body")
    default_css_class = 'card-body'

plugin_pool.register_plugin(BootstrapCardBodyPlugin)


class BootstrapCardFooterPlugin(TransparentContainer, CardChildBase):
    name = _("Card Footer")
    default_css_class = 'card-footer'

plugin_pool.register_plugin(BootstrapCardFooterPlugin)


class BootstrapCardPlugin(TransparentWrapper, BootstrapPluginBase):
    """
    Use this plugin to display a card with optional card-header and card-footer.
    """
    name = _("Card")
    default_css_class = 'card'
    require_parent = False
    parent_classes = ['BootstrapColumnPlugin']
    allow_children = True
    render_template = 'cascade/bootstrap4/card.html'

    @classmethod
    def get_identifier(cls, instance):
        try:
            return instance.card_header or instance.card_footer
        except AttributeError:
            pass
        return super().get_identifier(instance)

    @classmethod
    def get_child_classes(cls, slot, page, instance=None):
        """Restrict child classes of Card to one of each: Header, Body and Footer"""
        child_classes = super().get_child_classes(slot, page, instance)
        # allow only one child of type Header, Body, Footer
        for child in instance.get_children():
            if child.plugin_type in child_classes:
                child_classes.remove(child.plugin_type)
        return child_classes

plugin_pool.register_plugin(BootstrapCardPlugin)



import io
import json
import os

from cms.toolbar.utils import get_toolbar_from_request
from cms.plugin_rendering import StructureRenderer

from django import template
from django.conf import settings
from django.core.cache import caches
from django.template.exceptions import TemplateDoesNotExist
from django.contrib.staticfiles import finders
from django.utils.safestring import mark_safe
from classytags.arguments import Argument
from classytags.core import Options, Tag
from cmsplugin_cascade.strides import StrideContentRenderer

register = template.Library()


class StrideRenderer(Tag):
    """
    Render the serialized content of a placeholder field using the full cascade of plugins.
    {% render_cascade "cascade-data.json" %}

    Keyword arguments:
    datafile -- Filename containing the cascade tree. Must be file locatable by Django's
    static file finders.
    """
    name = 'render_cascade'
    options = Options(
        Argument('datafile'),
    )

    def render_tag(self, context, datafile):
        from sekizai.helpers import get_varname as get_sekizai_context_key
        from cmsplugin_cascade.strides import StrideContentRenderer

        cache = caches['default']
        tree_data_key = 'cascade-strides:' + datafile
        tree_data = cache.get(tree_data_key) if cache else None
        if tree_data is None:
            jsonfile = finders.find(datafile)
            if not jsonfile:
                raise IOError("Unable to find file: {}".format(datafile))

            with io.open(jsonfile) as fp:
                tree_data = json.load(fp)

        content_renderer = StrideContentRenderer(context['request'])
        with context.push(cms_content_renderer=content_renderer):
            content = content_renderer.render_cascade(context, tree_data)

        # some templates use Sekizai's templatetag `addtoblock` or `add_data`, which have to be re-added to the context
        cache = caches['default']
        if cache:
            sekizai_context_key = get_sekizai_context_key()
            SEKIZAI_CONTENT_HOLDER = cache.get_or_set(sekizai_context_key, context.get(sekizai_context_key))
            if SEKIZAI_CONTENT_HOLDER:
                for name in SEKIZAI_CONTENT_HOLDER:
                    context[sekizai_context_key][name] = SEKIZAI_CONTENT_HOLDER[name]
        return content

register.tag('render_cascade', StrideRenderer)


class RenderPlugin(Tag):
    """
    Alternative implementation of django-CMS's templatetag ``render_plugin``.
    It can either handle normal CMS plugins or Strides.
    """
    name = 'render_plugin'
    options = Options(
        Argument('plugin')
    )

    def render_tag(self, context, plugin):
        if not plugin:
            return ''

        request = context['request']
        toolbar = get_toolbar_from_request(request)
        if 'cms_content_renderer' in context and isinstance(context['cms_content_renderer'], StrideContentRenderer):
            content_renderer = context['cms_content_renderer']
        elif 'cms_renderer' in context:
            content_renderer = context['cms_renderer']
        elif 'cms_content_renderer' in context:
            content_renderer = context['cms_content_renderer']
        else:
            content_renderer = toolbar.content_renderer
        if isinstance(content_renderer, StructureRenderer):
            return content_renderer.render_plugin(plugin)
        else:
            return content_renderer.render_plugin(
                instance=plugin,
                context=context,
                editable=toolbar.edit_mode_active,
            )

register.tag('render_plugin', RenderPlugin)


@register.filter
def is_valid_image(image):
    try:
        return image.file.file
    except:
        return False


@register.simple_tag
def sphinx_docs_include(path):
    filename = os.path.join(settings.SPHINX_DOCS_ROOT, path)
    if not os.path.exists(filename):
        raise TemplateDoesNotExist("'{path}' does not exist".format(path=path))
    with io.open(filename) as fh:
        return mark_safe(fh.read())



from django.db import migrations

FIELD_MAPPINGS = {
    'BootstrapButtonPlugin': [
        ('button-type', 'button_type'),
        ('button-size', 'button_size'),
        ('button-options', 'button_options'),
        ('quick-float', 'quick_float'),
        ('icon-left', 'icon_left'),
        ('icon-right', 'icon_right'),
    ],
    'CarouselPlugin': [
        ('resize-options', 'resize_options'),
    ],
    'BootstrapGalleryPlugin': [
        ('image-shapes', 'image_shapes'),
        ('image-width-responsive', 'image_width_responsive'),
        ('image-width-fixed', 'image_width_fixed'),
        ('image-height', 'image_height'),
        ('thumbnail-width', 'thumbnail_width'),
        ('thumbnail-height', 'thumbnail_height'),
        ('resize-options', 'resize_options'),
    ],
    'BootstrapImagePlugin': [
        ('image-title', 'image_title'),
        ('alt-tag', 'alt_tag'),
        ('image-shapes', 'image_shapes'),
        ('image-width-responsive', 'image_width_responsive'),
        ('image-width-fixed', 'image_width_fixed'),
        ('image-height', 'image_height'),
        ('resize-options', 'resize_options'),
    ],
    'BootstrapJumbotronPlugin': [
        ('background-color', 'background_color'),
        ('background-repeat', 'background_repeat'),
        ('background-attachment', 'background_attachment'),
        ('background-vertical-position', 'background_vertical_position'),
        ('background-horizontal-position', 'background_horizontal_position'),
        ('background-size', 'background_size'),
        ('background-width-height', 'background_width_height'),
    ],
    'BootstrapPicturePlugin': [
        ('image-title', 'image_title'),
        ('alt-tag', 'alt_tag'),
        ('responsive-heights', 'responsive_heights'),
        ('responsive-zoom', 'responsive_zoom'),
        ('resize-options', 'resize_options'),
    ]
}


def forwards(apps, schema_editor):
    field_mappings = {}
    for key, maps in FIELD_MAPPINGS.items():
        field_mappings[key] = dict(maps)
    migrate_glossary(apps, field_mappings)


def backwards(apps, schema_editor):
    field_mappings = {}
    for key, maps in FIELD_MAPPINGS.items():
        field_mappings[key] = dict((m[1], m[0]) for m in maps)
    migrate_glossary(apps, field_mappings)


def migrate_glossary(apps, field_mappings):
    CascadeElement = apps.get_model('cmsplugin_cascade', 'CascadeElement')
    for element in CascadeElement.objects.all():
        if element.plugin_type not in field_mappings:
            continue
        glossary = dict(element.glossary)
        for srckey, value in element.glossary.items():
            dstkey = field_mappings[element.plugin_type].get(srckey)
            if dstkey and srckey in glossary:
                glossary[dstkey] = glossary.pop(srckey)
        element.glossary = glossary
        element.save()


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0013_iconfont'),
    ]

    operations = [
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



import os, io, json, shutil
from django.core.exceptions import SuspiciousFileOperation
from cmsplugin_cascade import app_settings

import tempfile
try:
    import czipfile as zipfile
except ImportError:
    import zipfile


def unzip_archive(label, zip_ref):
    common_prefix = os.path.commonprefix(zip_ref.namelist())
    if not common_prefix:
        raise SuspiciousFileOperation("The zip archive {} is not packed correctly".format(label))
    icon_font_root = app_settings.CMSPLUGIN_CASCADE['icon_font_root']
    try:
        try:
            os.makedirs(icon_font_root)
        except os.error:
            pass  # the directory exists already
        temp_folder = tempfile.mkdtemp(prefix='', dir=icon_font_root)
        for member in zip_ref.infolist():
            zip_ref.extract(member, temp_folder)
        font_folder = os.path.join(temp_folder, common_prefix)

        # this is specific to fontello
        with io.open(os.path.join(font_folder, 'config.json'), 'r') as fh:
            config_data = json.load(fh)
    except Exception as exc:
        shutil.rmtree(temp_folder, ignore_errors=True)
        raise SuspiciousFileOperation("Can not unzip uploaded archive {}: {}".format(label, exc))
    return os.path.relpath(font_folder, icon_font_root), config_data



import json, os
from docutils import nodes
from sphinx.builders.dirhtml import DirectoryHTMLBuilder
from sphinx.environment.adapters.toctree import TocTree


class FragmentsBuilder(DirectoryHTMLBuilder):
    name = 'fragments'

    def __init__(self, app):
        super().__init__(app)
        self.config.html_theme = 'bootstrap-fragments'
        self.config.html_theme_path.append(os.path.abspath(os.path.join(__file__, os.pardir, 'theme')))
        self.config.html_sidebars = {
            '**': ['globaltoc.html'],
        }
        self.docsmap_file = os.path.join(self.outdir, 'docsmap.json')
        if os.path.exists(self.docsmap_file):
            with open(self.docsmap_file, 'r') as fh:
                self.docs_map = json.load(fh, encoding='utf-8')
        else:
            self.docs_map = {}

    def prepare_writing(self, docnames):
        super().prepare_writing(docnames)
        for docname in docnames:
            doctree = self.env.get_doctree(docname)
            idx = doctree.first_child_matching_class(nodes.section)
            if idx is None or idx == -1:
                continue

            first_section = doctree[idx]
            idx = first_section.first_child_matching_class(nodes.title)
            if idx is None or idx == -1:
                continue

            doctitle = first_section[idx].astext()
            if doctitle:
                self.docs_map[docname] = doctitle
            self.globalcontext['toctree'] = lambda **kw: self._get_local_toctree(docname, **kw)

    def _get_local_toctree(self, docname, collapse=True, **kwds):
        # type: (str, bool, Any) -> str
        if 'includehidden' not in kwds:
            kwds['includehidden'] = False
        partials = TocTree(self.env).get_toctree_for(docname, self, collapse, **kwds)
        return self.render_partial(partials)['fragment']

    def finish(self):
        super().finish()
        with open(self.docsmap_file, 'w') as fh:
            json.dump(self.docs_map, fh)


def setup(app):
    app.require_sphinx('1.0')
    app.add_builder(FragmentsBuilder)



from django.db import migrations
from django.conf import settings


def forwards(apps, schema_editor):
    CascadePage = apps.get_model('cmsplugin_cascade', 'CascadePage')

    for cascade_page in CascadePage.objects.all():
        try:
            element_ids = cascade_page.glossary.pop('element_ids')
        except:
            continue
        cascade_page.glossary['element_ids'] = {settings.LANGUAGE_CODE: element_ids}
        cascade_page.save()


def backwards(apps, schema_editor):
    CascadePage = apps.get_model('cmsplugin_cascade', 'CascadePage')

    for cascade_page in CascadePage.objects.all():
        try:
            element_ids = cascade_page.glossary.pop('element_ids')
        except:
            continue
        cascade_page.glossary['element_ids'] = element_ids[settings.LANGUAGE_CODE]
        cascade_page.save()


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0029_json_field'),
    ]

    operations = [
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



from cmsplugin_cascade.bootstrap4.grid import Breakpoint
from cmsplugin_cascade.fields import MultiSizeField


class BootstrapMultiSizeField(MultiSizeField):
    """
    Some size input fields must be specified per Bootstrap breakpoint. Use this multiple
    input field to handle this.
    """
    def __init__(self, *args, **kwargs):
        properties = [bp.name for bp in Breakpoint]
        kwargs['sublabels'] = [bp.label for bp in Breakpoint]
        super().__init__(properties, *args, **kwargs)



from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from cms.plugin_pool import plugin_pool

from cmsplugin_cascade import app_settings
from cmsplugin_cascade.plugin_base import CascadePluginBase, TransparentContainer


class CustomSnippetPlugin(TransparentContainer, CascadePluginBase):
    """
    Allows to add a customized template anywhere. This plugins will be registered only if the
    project added a template using the configuration setting 'plugins_with_extra_render_templates'.
    """
    name = _("Custom Snippet")
    require_parent = False
    allow_children = True
    alien_child_classes = True
    render_template_choices = dict(app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_render_templates'].get('CustomSnippetPlugin', ()))
    render_template = 'cascade/generic/does_not_exist.html'  # default in case the template could not be found

    @classmethod
    def get_identifier(cls, instance):
        render_template = instance.glossary.get('render_template')
        if render_template:
            return format_html('{}', cls.render_template_choices.get(render_template))


if CustomSnippetPlugin.render_template_choices:
    # register only, if at least one template has been defined
    plugin_pool.register_plugin(CustomSnippetPlugin)



from django.forms import ChoiceField
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from cms.plugin_pool import plugin_pool
from entangled.forms import EntangledModelFormMixin

from cmsplugin_cascade.plugin_base import CascadePluginBase, TransparentContainer


class SimpleWrapperFormMixin(EntangledModelFormMixin):
    TAG_CHOICES = [(cls, _("<{}> – Element").format(cls)) for cls in ['div', 'span', 'section', 'article']] + \
                  [('naked', _("Naked Wrapper"))]

    tag_type = ChoiceField(
        choices=TAG_CHOICES,
        label=_("HTML element tag"),
        help_text=_('Choose a type for this HTML element.')
    )

    class Meta:
        entangled_fields = {'glossary': ['tag_type']}


class SimpleWrapperPlugin(TransparentContainer, CascadePluginBase):
    name = _("Simple Wrapper")
    parent_classes = None
    require_parent = False
    form = SimpleWrapperFormMixin
    allow_children = True
    alien_child_classes = True

    @classmethod
    def get_identifier(cls, instance):
        identifier = super().get_identifier(instance)
        tag_name = dict(SimpleWrapperFormMixin.TAG_CHOICES).get(instance.glossary.get('tag_type'))
        if tag_name:
            return format_html('{0} {1}', tag_name, identifier)
        return identifier

    def get_render_template(self, context, instance, placeholder):
        if instance.glossary.get('tag_type') == 'naked':
            return 'cascade/generic/naked.html'
        return 'cascade/generic/wrapper.html'

plugin_pool.register_plugin(SimpleWrapperPlugin)



from django.db import migrations
from cmsplugin_cascade.models import CascadeElement


def migrate_link(glossary):
    def foreign_key():
        return {'model': link['model'].lower(), 'pk': link['pk']}

    link = glossary.pop('link', None)
    if link:
        if link['type'] == 'cmspage':
            glossary.update({
                'link_type': 'cmspage',
                'cms_page': foreign_key(),
                'section': link['section'],
            })
        elif link['type'] == 'download':
            glossary.update({
                'link_type': 'download',
                'download_file': foreign_key(),
            })
        elif link['type'] == 'exturl':
            glossary.update({
                'link_type': 'exturl',
                'ext_url': link['url'],
            })
        elif link['type'] == 'email':
            glossary.update({
                'link_type': 'email',
                'mail_to': link['mail_to'],
            })
        elif link['type'] and link['type'] != 'none':
            glossary.update({
                'link_type': link['type'],
            })
        else:
            glossary.update({
                'link_type': '',
            })
        return True


def migrate_icon(glossary):
    icon_font = glossary.pop('icon_font', None)
    if icon_font and not isinstance(icon_font, dict):
        glossary.update({
            'icon_font': {'model': 'cmsplugin_cascade.iconfont', 'pk': int(icon_font)},
        })
        return True


def migrate_image(glossary):
    image = glossary.pop('image', None)
    if image:
        glossary.update({
            'image_file': {'model': 'filer.image', 'pk': image['pk']},
        })
        if 'width' in image and 'height' in image and 'exif_orientation' in image:
            glossary.update({
                '_image_properties': {'width': image['width'], 'height': image['height'],
                                      'exif_orientation': image['exif_orientation']},
            })
        return True


def forwards(apps, schema_editor):
    LINK_PLUGINS = ['TextLinkPlugin', 'LinkPlugin', 'BootstrapButtonPlugin', 'SimpleIconPlugin', 'FramedIconPlugin',
                    'BootstrapImagePlugin', 'BootstrapPicturePlugin', 'TextImagePlugin', 'TextIconPlugin']
    ICON_PLUGINS = ['BootstrapButtonPlugin', 'SimpleIconPlugin', 'FramedIconPlugin', 'TextIconPlugin']
    IMAGE_PLUGINS = ['BootstrapImagePlugin', 'BootstrapPicturePlugin', 'TextImagePlugin',
                     'BootstrapCarouselSlidePlugin', 'BootstrapJumbotronPlugin']
    for cascade_element in CascadeElement.objects.all():
        changed = False
        if cascade_element.plugin_type in LINK_PLUGINS:
            changed = migrate_link(cascade_element.glossary) or changed
        if cascade_element.plugin_type in ICON_PLUGINS:
            changed = migrate_icon(cascade_element.glossary) or changed
        if cascade_element.plugin_type in IMAGE_PLUGINS:
            changed = migrate_image(cascade_element.glossary) or changed
        if changed:
            cascade_element.save()


def backwards(apps, schema_editor):
    print("Backward migration not implemented")


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0026_cascadepage_menu_symbol'),
    ]

    operations = [
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0004_auto_20151112_0147'),
    ]

    operations = [
        migrations.CreateModel(
            name='CascadeClipboard',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('identifier', models.CharField(unique=True, max_length=50, verbose_name='Identifier')),
                ('data', models.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'verbose_name': 'Persited Clipboard Content',
                'verbose_name_plural': 'Persited Clipboard Content',
            },
        ),
    ]



from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool


@toolbar_pool.register
class CascadeClipboardToolbar(CMSToolbar):
    class Media:
        css = {
            'all': ['cascade/css/admin/clipboard.css']
        }



import json
from copy import deepcopy
from django import forms
from django.contrib.admin.helpers import AdminForm
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.apps import apps
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.models import SharedGlossary, CascadeElement
from .fields import SharedSettingsField


class SelectSharedGlossary(forms.Select):
    option_inherits_attrs = True

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        if value:
            attrs = {'data-glossary': json.dumps(self._get_data_glossary(value))}
        else:
            attrs = {}
        return super().create_option(name, value, label, selected, index, subindex, attrs)

    def _get_data_glossary(self, option_value):
        shared_instance = self.choices.queryset.get(pk=option_value)
        plugin_instance = plugin_pool.get_plugin(shared_instance.plugin_type)
        # use the saved glossary and filter it by fields marked as sharable
        glossary = dict((key, value) for key, value in shared_instance.glossary.items()
                        if key in plugin_instance.sharable_fields)
        self._enrich_link(glossary)
        return glossary

    def _enrich_link(self, glossary):
        """
        Enrich the dict glossary['link'] with an identifier onto the model
        """
        try:
            Model = apps.get_model(*glossary['link']['model'].split('.'))
            obj = Model.objects.get(pk=glossary['link']['pk'])
            glossary['link'].update(identifier=str(obj))
        except (KeyError, ObjectDoesNotExist):
            pass


class SharedFormMixin(EntangledModelFormMixin):
    """
    The change editor of plugins marked as sharable, are enriched by three fields:
        - a select box named 'shared_glossary',
        - a checkbox and a text input field
    these additional form fields are added during runtime.
    """
    shared_glossary = forms.ModelChoiceField(
        label=_("Shared Settings"),
        required=False,
        queryset=SharedGlossary.objects.all(),
        empty_label=_("Use individual settings"),
        help_text=_("Use settings shared with other plugins of this type"),
    )

    class Meta:
        untangled_fields = ['shared_glossary']


class SharableFormMixin(SharedFormMixin):
    save_settings_as = SharedSettingsField(
        label=_("Remember these settings as"),
    )

    class Meta:
        untangled_fields = ['save_settings_as']

    def clean_save_settings_as(self):
        identifier = self.cleaned_data['save_settings_as']
        if SharedGlossary.objects.filter(identifier=identifier).exclude(pk=self.instance.pk).exists():
            msg = _("The identifier '{}' has already been used, please choose another name.")
            raise ValidationError(msg.format(identifier))
        return identifier



class SharableGlossaryMixin(metaclass=forms.MediaDefiningClass):
    """
    Every plugin class of type ``CascadePluginBase`` additionally inherits from this mixin,
    if the plugin is marked as sharable.
    This class adds the appropriate methods to the plugin class in order to store
    an assortment of glossary values as a glossary reusable by other plugin instances.
    """
    ring_plugin = 'SharableGlossaryMixin'

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/sharableglossary.js']

    def changeform_view(self, request, object_id=None, form_url='', extra_context=None):
        if request.method == 'GET' and request.GET.get('glossary'):
            # Reached after the user changed the shared settings select field.
            # This recreates a form for the given plugin using the selected shared fields
            # as values for a faked object.
            extra_context = dict(extra_context or {})
            sharable_fields = {}
            try:
                shared_glossary = SharedGlossary.objects.get(id=request.GET.get('glossary'))
            except (ValueError, SharedGlossary.DoesNotExist):
                pass
            else:
                for field_name in self.sharable_fields:
                    shared_val = shared_glossary.glossary.get(field_name)
                    if shared_val:
                        sharable_fields[field_name] = shared_val

            if object_id:
                obj = self.get_object(request, object_id)
                glossary = deepcopy(obj.glossary)
                glossary.update(sharable_fields)
                # create fake object using values from shared fields to mimic the expected behaviour
                fake_obj = CascadeElement(glossary=glossary)
                ModelForm = self.get_form(request, fake_obj)
                form = ModelForm(instance=fake_obj)
            else:
                ModelForm = self.get_form(request)
                initial = self.get_changeform_initial_data(request)
                initial.update(sharable_fields)
                form = ModelForm(initial=initial)

            extra_context['adminform'] = AdminForm(
                form,
                list(self.get_fieldsets(request, obj)),
                {},
                [],
                model_admin=self)

        return super().changeform_view(request, object_id, form_url, extra_context)

    def get_form(self, request, obj=None, **kwargs):
        base_form = kwargs.pop('form', self.form)
        assert issubclass(base_form, EntangledModelFormMixin), "The Form class must inherit from EntangledModelFormMixin"
        if request.method == 'GET' and 'glossary' in request.GET:
            FormMixin = SharedFormMixin if request.GET['glossary'] else SharableFormMixin
        else:
            FormMixin = SharedFormMixin if obj and obj.shared_glossary else SharableFormMixin
        kwargs['form'] = type(base_form.__name__, (FormMixin, base_form), {})
        kwargs['form'].base_fields['shared_glossary'].limit_choices_to = dict(plugin_type=self.__class__.__name__)
        try:
            shared_glossary = SharedGlossary.objects.get(id=request.GET.get('glossary'))
        except ValueError:
            shared_glossary = None
        except SharedGlossary.DoesNotExist:
            shared_glossary = obj.shared_glossary if obj else None
        for field_name in self.sharable_fields:
            kwargs['form'].base_fields[field_name].disabled = bool(shared_glossary)
        return super().get_form(request, obj, **kwargs)

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        # in case checkbox for `save_settings_as` is checked, then we create a new entry in `models.SharedGlossary`
        # transferring the fields declared as sharable from this this plugin to that newly created object
        save_settings_as = form.cleaned_data.get('save_settings_as')
        if save_settings_as:
            # move data from form glossary to a SharedGlossary and refer to it
            shared_glossary, created = SharedGlossary.objects.get_or_create(
                plugin_type=self.__class__.__name__,
                identifier=save_settings_as,
            )
            assert created, "SharableCascadeForm.clean_save_settings_as() erroneously validated identifier '{}' " \
                            "as unique".format(save_settings_as)
            glry = form.cleaned_data['glossary']
            shared_glossary.glossary = dict((key, glry[key]) for key in self.sharable_fields if key in glry)
            shared_glossary.save()
            obj.shared_glossary = shared_glossary
            obj.save()

    @classmethod
    def get_data_representation(cls, instance):
        data = super().get_data_representation(instance)
        if instance.shared_glossary:
            data.setdefault('glossary', {})
            data['glossary'].update(instance.shared_glossary.glossary)
            data.update(shared_glossary=instance.shared_glossary.identifier)
        return data

    @classmethod
    def add_shared_reference(cls, instance, shared_glossary_identifier):
        try:
            shared_glossary = SharedGlossary.objects.get(plugin_type=instance.plugin_type,
                                                         identifier=shared_glossary_identifier)
        except SharedGlossary.DoesNotExist:
            pass
        else:
            instance.shared_glossary = shared_glossary
            instance.save(update_fields=['shared_glossary'])



import re

from django.conf import settings
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.forms import widgets
from django.forms.fields import BooleanField, CharField, ChoiceField, MultipleChoiceField
from django.forms.models import ModelForm
from django.http.response import HttpResponse
from django.template.loader import render_to_string
from django.urls import re_path
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _, gettext

from entangled.forms import EntangledModelForm

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade import app_settings
from cmsplugin_cascade.fields import SizeField
from cmsplugin_cascade.models import PluginExtraFields, TextEditorConfigFields, IconFont
from cmsplugin_cascade.extra_fields.mixins import ExtraFieldsMixin


class PluginExtraFieldsForm(EntangledModelForm):
    validation_pattern = re.compile(r'^[A-Za-z0-9_-]+$')

    class_names = CharField(
        label=_("CSS class names"),
        required=False,
        widget=widgets.TextInput(attrs={'style': 'width: 50em;'}),
        help_text=_("Freely selectable CSS classnames for this Plugin, separated by commas."),
    )

    multiple = BooleanField(
        label=_("Allow multiple"),
        required=False,
        help_text=_("Allow to select multiple of the above CSS classes."),
    )

    class Meta:
        untangled_fields = ['plugin_type', 'site']
        entangled_fields = {
            'css_classes': ['class_names', 'multiple'],
            'inline_styles': [],
        }

    def clean_class_names(self):
        value = self.cleaned_data['class_names']
        for val in value.split(','):
            val = val.strip()
            if val and not self.validation_pattern.match(val):
                msg = _("CSS class '{}' contains invalid characters.")
                raise ValidationError(msg.format(val))
        return value


class PluginExtraFieldsAdmin(admin.ModelAdmin):
    list_display = ['name', 'module', 'site', 'allowed_classes_styles']
    DISTANCE_UNITS = [
        ('px,em,rem,%', _("px, em, rem and %")),
        ('px,em,%', _("px, em and %")),
        ('px,rem,em', _("px, rem and em")),
        ('px,em', _("px and em")),
        ('px,rem,%', _("px, rem and %")),
        ('px,%', _("px and %")),
        ('px,rem', _("px and rem")),
        ('px', _("px")),
        ('%,rem', _("% and rem")),
        ('%', _("%")),
        ('px,em,rem,%,auto', _("px, em, rem, % and auto")),
        ('px,em,%,auto', _("px, em, % and auto")),
        ('px,rem,em,auto', _("px, rem, em and auto")),
        ('px,em,auto', _("px, em and auto")),
        ('px,rem,%,auto', _("px, rem, % and auto")),
        ('px,%,,auto', _("px, % and auto")),
        ('px,rem,auto', _("px, rem and auto")),
        ('px,auto', _("px and auto")),
        ('%,rem,auto', _("%, rem and auto")),
        ('%,auto', _("% and auto")),
    ]

    class Media:
        css = {'all': ('cascade/css/admin/partialfields.css',)}

    @cached_property
    def plugins_for_site(self):
        def show_in_backend(plugin):
            try:
                config = app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_fields'][plugin.__name__]
            except KeyError:
                return False
            else:
                assert issubclass(plugin, ExtraFieldsMixin), "Expected plugin to be of type `ExtraFieldsMixin`."
                return config.allow_override

        cascade_plugins = set([p for p in plugin_pool.get_all_plugins() if show_in_backend(p)])
        return [(p.__name__, '{}: {}'.format(p.module, str(p.name))) for p in cascade_plugins]

    def get_form(self, request, obj=None, **kwargs):
        form_fields = {
            'plugin_type': ChoiceField(choices=self.plugins_for_site),
        }
        for style, choices_tuples in app_settings.CMSPLUGIN_CASCADE['extra_inline_styles'].items():
            translated_style = gettext(style)
            form_fields['extra_fields:{0}'.format(style)] = MultipleChoiceField(
                label=_("Customized {0}:").format(translated_style),
                choices=[(c, c) for c in choices_tuples[0]],
                required=False,
                widget=widgets.CheckboxSelectMultiple,
                help_text=_("Allow these extra inlines styles for the given plugin type."),
            )
            if issubclass(choices_tuples[1], SizeField):
                form_fields['extra_units:{0}'.format(style)] = ChoiceField(
                    label=_("Units for {0}:").format(translated_style),
                    choices=self.DISTANCE_UNITS,
                    required=False,
                    help_text=_("Allow these size units for customized {0} fields.").format(translated_style),
                )
        inline_styles_fields = list(form_fields.keys())
        form = type('PluginExtraFieldsForm', (PluginExtraFieldsForm,), form_fields)
        form._meta.entangled_fields['inline_styles'] = inline_styles_fields
        kwargs.setdefault('form', form)
        return super().get_form(request, obj=None, **kwargs)

    def has_add_permission(self, request):
        """
        Only if at least one plugin uses the class ExtraFieldsMixin, allow to add an instance.
        """
        has_permission = super().has_add_permission(request)
        return has_permission and len(self.plugins_for_site) > 0

    def module(self, obj):
        return plugin_pool.get_plugin(obj.plugin_type).module
    module.short_description = _("Module")

    def allowed_classes_styles(self, obj):
        clsn = [cn for cn in obj.css_classes.get('class_names', '').split(',') if cn]
        sef = [len(group) for ef, group in obj.inline_styles.items() if ef.startswith('extra_fields:')]
        return "{} / {}".format(len(clsn), sum(sef))
    allowed_classes_styles.short_description = _("Allowed Classes and Styles")

admin.site.register(PluginExtraFields, PluginExtraFieldsAdmin)


class TextEditorConfigForm(ModelForm):
    validation_pattern = re.compile(r'^[A-Za-z0-9_-]+$')

    class Meta:
        fields = ['name', 'element_type', 'css_classes']

    def clean_css_classes(self):
        css_classes = []
        for val in self.cleaned_data['css_classes'].split(' '):
            if val:
                if self.validation_pattern.match(val.strip()):
                    css_classes.append(val)
                else:
                    raise ValidationError(_("'%s' is not a valid CSS class name.") % val)
        return ' '.join(css_classes)


class TextEditorConfigAdmin(admin.ModelAdmin):
    list_display = ['name', 'element_type']
    form = TextEditorConfigForm

    def get_urls(self):
        return [
            re_path(r'^wysiwig-config\.js$', self.render_texteditor_config,
                name='cascade_texteditor_config'),
        ] + super().get_urls()

    def render_texteditor_config(self, request):
        context = {
            'text_editor_configs': TextEditorConfigFields.objects.all(),
        }
        if 'cmsplugin_cascade.icon' in settings.INSTALLED_APPS:
            context['icon_fonts'] = IconFont.objects.all()
        javascript = render_to_string('cascade/admin/ckeditor.wysiwyg.txt', context)
        return HttpResponse(javascript, content_type='application/javascript')

admin.site.register(TextEditorConfigFields, TextEditorConfigAdmin)



from django.forms import ChoiceField
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.fields import SizeField, ColorField, BorderChoiceField
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin
from cmsplugin_cascade.icon.forms import IconFormMixin
from cmsplugin_cascade.icon.plugin_base import IconPluginMixin


class FramedIconFormMixin(IconFormMixin):
    SIZE_CHOICES = [('{}em'.format(c), "{} em".format(c)) for c in range(1, 13)]

    RADIUS_CHOICES = [(None, _("Square"))] + \
        [('{}px'.format(r), "{} px".format(r)) for r in (1, 2, 3, 5, 7, 10, 15, 20)] + \
        [('50%', _("Circle"))]

    TEXT_ALIGN_CHOICES = [
        (None, _("Do not align")),
        ('text-left', _("Left")),
        ('text-center', _("Center")),
        ('text-right', _("Right"))
    ]

    font_size = SizeField(
        label=_("Icon size"),
        allowed_units=['px', 'em'],
        initial='1em',
    )

    color = ColorField(
        label=_("Icon color"),
    )

    background_color = ColorField(
        label=_("Background color"),
        inherit_color=True,
    )

    text_align = ChoiceField(
        choices=TEXT_ALIGN_CHOICES,
        label=_("Text alignment"),
        required=False,
        help_text=_("Align the icon inside the parent column.")
    )

    border = BorderChoiceField(
        label=_("Set border"),
    )

    border_radius = ChoiceField(
        choices=RADIUS_CHOICES,
        label=_("Border radius"),
        required=False,
    )

    class Meta:
        entangled_fields = {'glossary': ['font_size', 'color', 'background_color', 'text_align', 'border',
                                         'border_radius']}


class FramedIconPlugin(IconPluginMixin, LinkPluginBase):
    name = _("Icon with frame")
    parent_classes = None
    require_parent = False
    allow_children = False
    render_template = 'cascade/bootstrap4/framedicon.html'
    model_mixins = (LinkElementMixin,)
    form = type('FramedIconForm', (LinkFormMixin, FramedIconFormMixin), {'require_link': False})
    ring_plugin = 'FramedIconPlugin'

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/framediconplugin.js']

    @classmethod
    def get_tag_type(self, instance):
        if instance.glossary.get('text_align') or instance.glossary.get('font_size'):
            return 'div'

    @classmethod
    def get_css_classes(cls, instance):
        css_classes = cls.super(FramedIconPlugin, cls).get_css_classes(instance)
        text_align = instance.glossary.get('text_align')
        if text_align:
            css_classes.append(text_align)
        return css_classes

    @classmethod
    def get_inline_styles(cls, instance):
        inline_styles = cls.super(FramedIconPlugin, cls).get_inline_styles(instance)
        inline_styles['font-size'] = instance.glossary.get('font_size', '1em')
        return inline_styles

    def render(self, context, instance, placeholder):
        context = self.super(FramedIconPlugin, self).render(context, instance, placeholder)
        styles = {'display': 'inline-block'}
        color, inherit = instance.glossary.get('color', (ColorField.DEFAULT_COLOR, True))
        if not inherit:
            styles['color'] = color
        background_color, inherit = instance.glossary.get('background_color', (ColorField.DEFAULT_COLOR, True))
        if not inherit:
            styles['background-color'] = background_color
        border = instance.glossary.get('border')
        if isinstance(border, list) and border[0] and border[1] != 'none':
            styles.update(border='{0} {1} {2}'.format(*border))
            radius = instance.glossary.get('border_radius')
            if radius:
                styles['border-radius'] = radius
        attrs = []
        if 'icon_font_class' in context:
            attrs.append(format_html('class="{}"', context['icon_font_class']))
        attrs.append(format_html('style="{}"', format_html_join('', '{0}:{1};', [(k, v) for k, v in styles.items()])))
        context['icon_font_attrs'] = mark_safe(' '.join(attrs))
        return context

plugin_pool.register_plugin(FramedIconPlugin)



from django.conf import settings
from django.utils.translation import gettext_lazy as _
from cms.extensions.toolbar import ExtensionToolbar
from cms.toolbar_pool import toolbar_pool
from cms.toolbar.items import Break
from cms.cms_toolbars import PAGE_MENU_SECOND_BREAK
from cmsplugin_cascade.models import CascadePage


class CascadePageToolbar(ExtensionToolbar):
    model = CascadePage

    def populate(self):
        current_page_menu = self._setup_extension_toolbar()
        if current_page_menu:
            # retrieves the instance of the current extension (if any) and the toolbar item URL
            page_extension, url = self.get_page_extension_admin()
            if url:
                position = current_page_menu.find_first(Break, identifier=PAGE_MENU_SECOND_BREAK)
                disabled = not self.toolbar.edit_mode_active
                current_page_menu.add_modal_item(_("Extra Page Fields"), position=position, url=url, disabled=disabled)

if settings.CMSPLUGIN_CASCADE['register_page_editor']:
    toolbar_pool.register(CascadePageToolbar)



from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar
from cmsplugin_cascade import app_settings


@toolbar_pool.register
class SegmentationToolbar(CMSToolbar):
    def populate(self):
        if not self.request.user.has_perm('cmsplugin_cascade.view_segmentation'):
            return
        menu = self.toolbar.get_or_create_menu('segmentation', _("Segmentation"))
        for sgm in app_settings.CMSPLUGIN_CASCADE['segmentation_mixins']:
            SegmentationMixin = import_string(sgm[1])
            populate_handler = getattr(SegmentationMixin, 'populate_toolbar', None)
            if callable(populate_handler):
                populate_handler(menu, self.request)



import html

from django.core.exceptions import ValidationError
from django.forms.fields import CharField, ChoiceField
from django.utils.translation import gettext_lazy as _
from django.utils.html import format_html
from django.template import engines, TemplateSyntaxError, Template as DjangoTemplate, Context as TemplateContext
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.plugin_base import CascadePluginBase, TransparentContainer


class Template(DjangoTemplate):
    def render(self, context):
        if isinstance(context, dict):
            context = TemplateContext(context)
        return super().render(context)


class SegmentFormMixin(EntangledModelFormMixin):
    eval_template_string = '{{% if {} %}}True{{% endif %}}'

    open_tag = ChoiceField()  # redeclared in get_form

    condition = CharField(
        label=_("Condition evaluation"),
        required=False,
        # widget=widgets.TextInput(attrs={'width': '100%'}),
        help_text=_("Evaluation as used in Django's template tags for conditions")
    )

    class Meta:
        entangled_fields = {'glossary': ['open_tag', 'condition']}

    def __init__(self, *args, **kwargs):
        if 'instance' in kwargs and kwargs['instance']:
            pass
        super().__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data['open_tag'] in ('if', 'elif'):
            if not cleaned_data['condition']:
                raise ValidationError(_("The evaluation condition is missing or empty."))
            try:
                condition = html.unescape(cleaned_data['condition'])
                engines['django'].from_string(self.eval_template_string.format(condition))
            except TemplateSyntaxError as err:
                raise ValidationError(_("Unable to evaluate condition: {}").format(str(err)))
        elif cleaned_data['open_tag'] == 'else':
            cleaned_data['condition'] = ''  # empty condition for else-block
        return cleaned_data


class SegmentPlugin(TransparentContainer, CascadePluginBase):
    """
    A Segment is a part of the DOM which is rendered or not, depending on a condition.
    As condition you may use any expression which is valid inside a Django's template tag,
    such as ``{% if ... %}``.
    """
    name = _("Segment")
    default_template = engines['django'].from_string('{% load cms_tags %}{% for plugin in instance.child_plugin_instances %}{% render_plugin plugin %}{% endfor %}')
    hiding_template_string = '{% load cms_tags %}<div style="display: none;">{% for plugin in instance.child_plugin_instances %}{% render_plugin plugin %}{% endfor %}</div>'
    hiding_template = engines['django'].from_string(hiding_template_string)
    debug_error_template = '<!-- segment condition "{condition}" for plugin: {instance_id} failed: "{message}" -->{template_string}'
    empty_template = engines['django'].from_string('{% load l10n %}<!-- segment condition for plugin: {{ instance.id|unlocalize }} did not evaluate -->')
    ring_plugin = 'SegmentPlugin'
    require_parent = False
    direct_parent_classes = None
    allow_children = True
    child_classes = None
    cache = False

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/segmentplugin.js']

    @classmethod
    def get_identifier(cls, obj):
        try:
            return format_html("<strong><em>{open_tag}</em></strong> {condition}", **obj.glossary)
        except KeyError:
            return ''

    def get_render_template(self, context, instance, placeholder):
        def conditionally_eval():
            condition = html.unescape(instance.glossary['condition'])
            evaluated_to = False
            template_error_message = None
            try:
                template_string = SegmentFormMixin.eval_template_string.format(condition)
                eval_template = engines['django'].from_string(template_string)
                evaluated_to = eval_template.render(context) == 'True'
            except TemplateSyntaxError as err:
                template_error_message = str(err)
            finally:
                if evaluated_to:
                    request._evaluated_segments[instance.pk] = True
                    template = self.default_template
                else:
                    request._evaluated_segments[instance.pk] = False
                    if edit_mode:
                        # In edit mode, hidden plugins have to be rendered nevertheless. Therefore
                        # we use `style="display: none"`, otherwise the plugin would be invisible
                        # in structure mode, while editing.
                        if template_error_message:
                            template = self.debug_error_template.format(condition=condition,
                                instance_id=instance.pk, message=template_error_message,
                                template_string=self.hiding_template_string)
                            template = engines['django'].from_string(template)
                        else:
                            template = self.hiding_template
                    else:
                        template = self.empty_template
            return template

        request = context['request']
        edit_mode = self.in_edit_mode(request, placeholder)
        open_tag = instance.glossary.get('open_tag')
        if open_tag == 'if':
            template = conditionally_eval()
        else:
            assert open_tag in ('elif', 'else'), "Expected openening templatetag to be 'elif' or 'else'."
            prev_instance = self.get_previous_instance(instance)
            if prev_instance is None:
                # this can happen, if one moved an `else`- or `elif`-segment in front of an `if`-segment
                template = edit_mode and self.hiding_template or self.empty_template
            elif request._evaluated_segments.get(prev_instance.pk):
                request._evaluated_segments[instance.pk] = (open_tag == 'elif')
                # in edit mode hidden plugins have to be rendered nevertheless
                template = edit_mode and self.hiding_template or self.empty_template
            elif open_tag == 'elif':
                template = conditionally_eval()
            else:
                template = self.default_template
        return template

    def render(self, context, instance, placeholder):
        request = context['request']
        if not hasattr(request, '_evaluated_segments'):
            request._evaluated_segments = {}
        context.update(instance.get_context_override(request))
        return self.super(SegmentPlugin, self).render(context, instance, placeholder)

    def get_form(self, request, obj=None, **kwargs):
        choices = [('if', _("if"))]
        if obj is None or self._get_previous_open_tag(obj) in ('if', 'elif'):
            # if obj is None, we are adding a SegmentPlugin, hence we must offer all conditional
            # tags, which however may be rectified during the save() operation
            choices.extend([('elif', _("elif")), ('else', _("else"))])
        open_tag = ChoiceField(
            label=_("Condition tag"),
            choices=choices,
            help_text=_("Django's condition tag")
        )
        form = type('SegmentForm', (SegmentFormMixin,), {'open_tag': open_tag})
        kwargs.setdefault('form', form)
        return super().get_form(request, obj, **kwargs)

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if obj.glossary['open_tag'] != 'if' and self._get_previous_open_tag(obj) not in ('if', 'elif'):
           # rectify to an `if`-segment, when plugin is not saved next to an `if`- or `elif`-segment
           obj.glossary['open_tag'] = 'if'
           obj.save(update_fields=['glossary'])

    def _get_previous_open_tag(self, obj):
        """
        Return the open tag of the previous sibling
        """
        prev_instance = self.get_previous_instance(obj)
        if prev_instance and prev_instance.plugin_type == self.__class__.__name__:
            return prev_instance.glossary.get('open_tag')

plugin_pool.register_plugin(SegmentPlugin)



from django.forms import widgets, CharField, ModelChoiceField
from django.utils.translation import gettext_lazy as _

from cmsplugin_cascade.models import IconFont

from entangled.forms import EntangledModelFormMixin


def get_default_icon_font():
    try:
        return IconFont.objects.get(is_default=True).id
    except IconFont.DoesNotExist:
        return ''


class IconFormMixin(EntangledModelFormMixin):
    icon_font = ModelChoiceField(
        IconFont.objects.all(),
        label=_("Font"),
        initial=get_default_icon_font,
    )

    symbol = CharField(
        widget=widgets.HiddenInput(),
        label=_("Select Symbol"),
    )

    class Meta:
        entangled_fields = {'glossary': ['icon_font', 'symbol']}

    def __init__(self, *args, **kwargs):
        if not getattr(self, 'require_icon', True):
            self.declared_fields['icon_font'].required = False
            self.declared_fields['icon_font'].empty_label = _("No Icon")
            self.declared_fields['icon_font'].initial = None
            self.declared_fields['symbol'].required = False
        super().__init__(*args, **kwargs)



from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import ObjectDoesNotExist
from django.forms import MediaDefiningClass, widgets
from django.forms.fields import CharField, ChoiceField, MultipleChoiceField
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cmsplugin_cascade import app_settings
from cmsplugin_cascade.fields import SizeField


class ExtraFieldsMixin(metaclass=MediaDefiningClass):
    """
    If a Cascade plugin is listed in ``settings.CMSPLUGIN_CASCADE['plugins_with_extra_fields']``,
    then this ``ExtraFieldsMixin`` class is added automatically to its plugin class in order to
    offer extra fields for customizing CSS classes and styles.
    """

    def __str__(self):
        return self.plugin_class.get_identifier(self)

    def get_form(self, request, obj=None, **kwargs):
        from cmsplugin_cascade.models import PluginExtraFields
        from .config import PluginExtraFieldsConfig

        clsname = self.__class__.__name__
        try:
            site = get_current_site(request)
            extra_fields = PluginExtraFields.objects.get(plugin_type=clsname, site=site)
        except ObjectDoesNotExist:
            extra_fields = app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_fields'].get(clsname)

        if isinstance(extra_fields, (PluginExtraFields, PluginExtraFieldsConfig)):
            form_fields = {}

            # add a text input field to let the user name an ID tag for this HTML element
            if extra_fields.allow_id_tag:
                form_fields['extra_element_id'] = CharField(
                    label=_("Named Element ID"),
                )

            # add a select box to let the user choose one or more CSS classes
            class_names, choices = extra_fields.css_classes.get('class_names'), None
            if isinstance(class_names, (list, tuple)):
                choices = [(clsname, clsname) for clsname in class_names]
            elif isinstance(class_names, str):
                choices = [(clsname, clsname) for clsname in class_names.replace(' ', ',').split(',') if clsname]
            if choices:
                if extra_fields.css_classes.get('multiple'):
                    form_fields['extra_css_classes'] = MultipleChoiceField(
                        label=_("Customized CSS Classes"),
                        choices=choices,
                        required=False,
                        widget=widgets.CheckboxSelectMultiple,
                        help_text=_("Customized CSS classes to be added to this element."),
                    )
                else:
                    choices.insert(0, (None, _("Select CSS")))
                    form_fields['extra_css_classes'] = ChoiceField(
                        label=_("Customized CSS Class"),
                        choices=choices,
                        required=False,
                        help_text=_("Customized CSS class to be added to this element."),
                    )

            # add input fields to let the user enter styling information
            for style, choices_list in app_settings.CMSPLUGIN_CASCADE['extra_inline_styles'].items():
                inline_styles = extra_fields.inline_styles.get('extra_fields:{0}'.format(style))
                if not inline_styles:
                    continue
                Field = choices_list[1]
                for inline_style in inline_styles:
                    key = 'extra_inline_styles:{0}'.format(inline_style)
                    field_kwargs = {
                        'label': '{0}: {1}'.format(style, inline_style),
                        'required': False,
                    }
                    if issubclass(Field, SizeField):
                        field_kwargs['allowed_units'] = extra_fields.inline_styles.get('extra_units:{0}'.format(style)).split(',')
                    form_fields[key] = Field(**field_kwargs)

            # extend the form with some extra fields
            base_form = kwargs.pop('form', self.form)
            assert issubclass(base_form, EntangledModelFormMixin), "Form must inherit from EntangledModelFormMixin"
            class Meta:
                entangled_fields = {'glossary': list(form_fields.keys())}
            form_fields['Meta'] = Meta
            kwargs['form'] = type(base_form.__name__, (base_form,), form_fields)
        return super().get_form(request, obj, **kwargs)

    @classmethod
    def get_css_classes(cls, obj):
        """Enrich list of CSS classes with customized ones"""
        css_classes = super().get_css_classes(obj)
        extra_css_classes = obj.glossary.get('extra_css_classes')
        if extra_css_classes:
            if isinstance(extra_css_classes, str):
                css_classes.append(extra_css_classes)
            elif isinstance(extra_css_classes, (list, tuple)):
                css_classes.extend(extra_css_classes)
        return css_classes

    @classmethod
    def get_inline_styles(cls, obj):
        """Enrich inline CSS styles with customized ones"""
        inline_styles = super().get_inline_styles(obj)
        extra_inline_styles = app_settings.CMSPLUGIN_CASCADE['extra_inline_styles']
        for key, eis in obj.glossary.items():
            if key.startswith('extra_inline_styles:'):
                _, prop = key.split(':')
                if isinstance(eis, dict):
                    # a multi value field, storing values as dict
                    inline_styles.update(dict((k, v) for k, v in eis.items() if v))
                elif isinstance(eis, (list, tuple)):
                    # a multi value field, storing values as list
                    for props, field_class in extra_inline_styles.values():
                        if prop in props:
                            inline_styles.update({prop: field_class.css_value(eis)})
                            break
                elif isinstance(eis, str):
                    inline_styles.update({prop: eis})
        return inline_styles

    @classmethod
    def get_html_tag_attributes(cls, obj):
        attributes = super().get_html_tag_attributes(obj)
        extra_element_id = obj.glossary.get('extra_element_id')
        if extra_element_id:
            attributes.update(id=extra_element_id)
        return attributes

    @classmethod
    def get_identifier(cls, obj):
        identifier = super().get_identifier(obj)
        extra_element_id = obj.glossary and obj.glossary.get('extra_element_id')
        if extra_element_id:
            return format_html('{0}<em>{1}:</em> ', identifier, extra_element_id)
        return identifier



from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.plugin_base import CascadePluginBase


class HorizontalRulePlugin(CascadePluginBase):
    name = _("Horizontal Rule")
    parent_classes = None
    allow_children = False
    tag_type = 'hr'
    render_template = 'cascade/generic/single.html'

plugin_pool.register_plugin(HorizontalRulePlugin)



import io
import json
import os
from django.conf import settings
from django.forms import fields
from django.utils.translation import gettext_lazy as _
from django_select2.forms import Select2Widget
from cms.models.pagemodel import Page
from cmsplugin_cascade.link.plugin_base import LinkPluginBase
from cmsplugin_cascade.link.forms import LinkForm


class DocumentationSelect2Widget(Select2Widget):
    def render(self, name, value, attrs=None, renderer=None):
        html = super().render(name, value, attrs=attrs, renderer=None)
        return html


def get_documents_map():
    docsmap_file = os.path.join(settings.SPHINX_DOCS_ROOT, 'docsmap.json')
    if not os.path.exists(docsmap_file):
        return ()
    with io.open(docsmap_file) as fh:
        docs_map = json.load(fh, encoding='utf-8')
    result = []
    for path, title in docs_map.items():
        bits = path.split('/')
        if len(bits) == 2 and bits[1] == 'index':
            result.append((bits[0], title))
        elif bits[0] != 'index':
            result.append((path, title))
    return result


class SphinxDocsLinkForm(LinkForm):
    LINK_TYPE_CHOICES = [
        ('cmspage', _("CMS Page")),
        ('documentation', _("Documentation")),
        ('exturl', _("External URL")),
        ('email', _("Mail To")),
    ]

    documentation = fields.ChoiceField(
        required=False,
        label='',
        choices=get_documents_map(),
        widget=Select2Widget,
        help_text=_("An internal link onto a documentation page"),
    )

    def clean_documentation(self):
        if self.cleaned_data.get('link_type') == 'documentation':
            self.cleaned_data['link_data'] = {
                'type': 'documentation',
                'value': self.cleaned_data.get('documentation'),
            }

    def set_initial_documentation(self, initial):
        try:
            initial['documentation'] = initial['link']['value']
        except KeyError:
            pass


class SphinxDocsLinkPlugin(LinkPluginBase):
    fields = [
        ('link_type', 'cms_page', 'section', 'documentation', 'ext_url', 'mail_to'),
        'glossary',
    ]
    ring_plugin = 'SphinxDocsLinkPlugin'

    class Media:
        js = ['cascade/sphinx/js/link_plugin.js']

    @classmethod
    def get_link(cls, obj):
        link = obj.glossary.get('link', {})
        if link.get('type') == 'documentation':
            page = Page.objects.filter(navigation_extenders='DocumentationMenu', publisher_is_draft=False).first()
            if page:
                return page.get_public_url()
        return super().get_link(obj)



# Generated by Django 4.0.6.dev20220601124058 on 2022-08-02 08:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0030_separate_ids_per_language'),
    ]

    operations = [
        migrations.AlterField(
            model_name='texteditorconfigfields',
            name='element_type',
            field=models.CharField(choices=[('p', 'p'), ('h1', 'h1'), ('h2', 'h2'), ('h3', 'h3'), ('h4', 'h4'), ('h5', 'h5'), ('h6', 'h6'), ('pre', 'pre'), ('address', 'address'), ('div', 'div'), ('span', 'span'), ('ol', 'ol'), ('ul', 'ul')], max_length=12, verbose_name='Element Type'),
        ),
    ]



from cmsplugin_cascade.models import CascadeModelBase
from cmsplugin_cascade.plugin_base import fake_proxy_models
from django.db import migrations
from django.utils.functional import classproperty


class Migration(migrations.Migration):
    """
    This migration is a noop. It pretends that the proxy models created by Cascade do already
    exists, so that ``./manage.py makemigrations`` does not create any migration file for you.
    """

    dependencies = [
        ('cmsplugin_cascade', '0016_shared_glossary_uneditable'),
    ]

    @classproperty
    def operations(self):
        for name, bases in sorted(fake_proxy_models.items()):
            bases = tuple(
                b._meta.app_label + '.' + b._meta.model_name if issubclass(b, CascadeModelBase) else b
                for b in bases
            )
            yield migrations.CreateModel(
                name=name,
                fields=[],
                options={'proxy': True},
                bases=bases,
            )



from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.forms import TextLinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin


class TextLinkPlugin(LinkPluginBase):
    name = _("Link")
    model_mixins = (LinkElementMixin,)
    text_enabled = True
    render_template = 'cascade/link/text-link.html'
    ring_plugin = 'TextLinkPlugin'
    form = type('TextLinkForm', (LinkFormMixin, TextLinkFormMixin), {})
    parent_classes = ['TextPlugin']

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/textlinkplugin.js']

    @classmethod
    def get_identifier(cls, obj):
        return mark_safe(obj.glossary.get('link_content', ''))

    @classmethod
    def requires_parent_plugin(cls, slot, page):
        """
        Workaround for `PluginPool.get_all_plugins()`, otherwise TextLinkPlugin is not allowed
        as a child of a `TextPlugin`.
        """
        return False

plugin_pool.register_plugin(TextLinkPlugin)



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0015_carousel_slide'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cascadeelement',
            name='shared_glossary',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='cmsplugin_cascade.SharedGlossary'),
        ),
        migrations.AlterModelOptions(
            name='cascadeelement',
            options={'verbose_name': 'Element'},
        ),
    ]



import re
import json
import warnings

from django.apps import apps
from django.core.exceptions import ValidationError
from django.db.models.fields.related import ManyToOneRel
from django.forms import widgets
from django.forms.fields import Field, CharField, ChoiceField, BooleanField, MultiValueField
from django.forms.utils import ErrorList
from django.core.validators import ProhibitNullCharactersValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _, pgettext

from cmsplugin_cascade import app_settings
from cmsplugin_cascade.widgets import ColorPickerWidget, BorderChoiceWidget, MultipleTextInputWidget
from filer.fields.image import FilerImageField, AdminImageFormField
from filer.settings import settings as filer_settings


class GlossaryField:
    """
    Deprecated.
    Behave similar to django.forms.Field, encapsulating a partial dictionary, stored as
    JSONField in the database.
    """
    def __init__(self, widget, label=None, name=None, initial='', help_text='', error_class=ErrorList):
        warnings.warn("GlossaryField is deprecated")
        self.name = name
        if not isinstance(widget, widgets.Widget):
            raise AttributeError('`widget` must inherit from django.forms.widgets.Widget')
        if label is None:
            label = name
        self.widget = widget
        self.label = label
        self.initial = initial
        self.help_text = help_text
        self.error_class = error_class

    def run_validators(self, value):
        if not callable(getattr(self.widget, 'validate', None)):
            return
        errors = []
        if callable(getattr(self.widget, '__iter__', None)):
            for field_name in self.widget:
                try:
                    self.widget.validate(value.get(self.name), field_name)
                except ValidationError as e:
                    if isinstance(getattr(e, 'params', None), dict):
                        e.params.update(label=self.label)
                    messages = self.error_class([m for m in e.messages])
                    errors.extend(messages)
        else:
            try:
                self.widget.validate(value.get(self.name))
            except ValidationError as e:
                if isinstance(getattr(e, 'params', None), dict):
                    e.params.update(label=self.label)
                errors = self.error_class([m for m in e.messages])
        if errors:
            raise ValidationError(errors)

    def get_element_ids(self, prefix_id):
        """
        Returns a single or a list of element ids, one for each input widget of this field
        """
        if isinstance(self.widget, widgets.MultiWidget):
            ids = ['{0}_{1}_{2}'.format(prefix_id, self.name, field_name) for field_name in self.widget]
        elif isinstance(self.widget, (widgets.SelectMultiple, widgets.RadioSelect)):
            ids = ['{0}_{1}_{2}'.format(prefix_id, self.name, k) for k in range(len(self.widget.choices))]
        else:
            ids = ['{0}_{1}'.format(prefix_id, self.name)]
        return ids


class BorderChoiceField(MultiValueField):
    BORDER_STYLES = ['none', 'solid', 'dashed', 'dotted', 'double', 'groove', 'hidden',
                     'inset', 'outset', 'ridge']

    def __init__(self, *args, **kwargs):
        choices = [(s, s) for s in self.BORDER_STYLES]
        with_alpha = kwargs.pop('with_alpha', app_settings.CMSPLUGIN_CASCADE['color_picker_with_alpha'])
        widget = kwargs.pop('widget', BorderChoiceWidget(choices, with_alpha))
        fields = [
            SizeField(),
            ChoiceField(choices=choices),
            CharField(),
        ]
        kwargs['initial'] = ['0px', 'none', '#000000']
        super().__init__(fields=fields, widget=widget, *args, **kwargs)

    def prepare_value(self, value):
        return value

    def compress(self, data_list):
        return data_list

    @classmethod
    def css_value(self, values):
        return ' '.join(values)


class SelectTextAlignField(ChoiceField):
    CHOICES = [('left', 'left'), ('center', 'center'), ('right', 'right'), ('justify', 'justify')]

    def __init__(self, *args, **kwargs):
        super().__init__(choices=self.CHOICES, *args, **kwargs)


class SelectOverflowField(ChoiceField):
    CHOICES = [('auto', 'auto'), ('scroll', 'scroll'), ('hidden', 'hidden')]

    def __init__(self, *args, **kwargs):
        super().__init__(choices=self.CHOICES, *args, **kwargs)


@deconstructible
class ColorValidator():
    message = _("'%(color)s' is not a valid color code.")
    code = 'invalid_color_code'

    def __init__(self, with_alpha):
        if with_alpha:
            self.validation_pattern = re.compile(r'(#(?:[0-9a-fA-F]{2}){2,4}|(#[0-9a-fA-F]{3})|(rgb|hsl)a?\((-?\d+%?[,\s]+){2,3}\s*[\d\.]+%?\))')
        else:
            self.validation_pattern = re.compile(r'(#(?:[0-9a-fA-F]{2}){2,3}|(#[0-9a-fA-F]{3})|(rgb|hsl))')

    def __call__(self, value):
        color, inherit = value
        match = self.validation_pattern.match(color)
        if not (inherit or match):
            params = {'color': color}
            raise ValidationError(self.message, code=self.code, params=params)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.validation_pattern == other.validation_pattern and
            self.message == other.message and
            self.code == other.code
        )


class ColorField(MultiValueField):
    DEFAULT_COLOR = '#808080'

    def __init__(self, inherit_color=False, default_color=DEFAULT_COLOR, *args, **kwargs):
        kwargs.pop('required', None)
        with_alpha = kwargs.pop('with_alpha', app_settings.CMSPLUGIN_CASCADE['color_picker_with_alpha'])
        widget = kwargs.pop('widget', ColorPickerWidget(with_alpha))
        fields = [
            CharField(initial=default_color),
            BooleanField(initial=inherit_color, required=False),
        ]
        kwargs['initial'] = [default_color, inherit_color]
        super().__init__(fields=fields, widget=widget, *args, **kwargs)
        self.validators.append(ColorValidator(with_alpha))
        self.validators.append(ProhibitNullCharactersValidator())

    def compress(self, data_list):
        self.run_validators(data_list)
        return data_list

    @classmethod
    def css_value(self, values):
        return values[0]


@deconstructible
class SizeUnitValidator():
    allowed_units = []
    message = _("'%(value)s' is not a valid size unit. Allowed units are: %(allowed_units)s.")
    code = 'invalid_size_unit'

    def __init__(self, allowed_units=None, allow_negative=True):
        possible_units = ['rem', 'px', 'em', '%', 'auto']
        if allowed_units is None:
            self.allowed_units = possible_units
        else:
            self.allowed_units = [au for au in allowed_units if au in possible_units]
        units_with_value = list(self.allowed_units)
        if 'auto' in self.allowed_units:
            self.allow_auto = True
            units_with_value.remove('auto')
        else:
            self.allow_auto = False
        if allow_negative:
            patterns = r'^(-?\d+(\.\d+)?)({})$'.format('|'.join(units_with_value))
        else:
            patterns = r'^(\d+(\.\d+)?)({})$'.format('|'.join(units_with_value))
        self.validation_pattern = re.compile(patterns)

    def __call__(self, value):
        if self.allow_auto and value == 'auto':
            return
        match = self.validation_pattern.match(value)
        try:
            float(match.group(1))
        except (AttributeError, ValueError):
            allowed_units = " {} ".format(pgettext('allowed_units', "or")).join("'{}'".format(u) for u in self.allowed_units)
            params = {'value': value, 'allowed_units': allowed_units}
            raise ValidationError(self.message, code=self.code, params=params)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.allowed_units == other.allowed_units and
            self.message == other.message and
            self.code == other.code
        )


class SizeField(Field):
    """
    Use this field for validating input containing a value ending in ``px``, ``em``, ``rem`` or ``%``.
    Use it for values representing a size, margin, padding, width or height.
    """
    def __init__(self, *, allowed_units=None, **kwargs):
        self.empty_value = ''
        super().__init__(**kwargs)
        self.validators.append(SizeUnitValidator(allowed_units))
        self.validators.append(ProhibitNullCharactersValidator())

    def to_python(self, value):
        """Return a stripped string."""
        if value not in self.empty_values:
            value = str(value).strip()
        if value in self.empty_values:
            return self.empty_value
        return value


class MultiSizeField(MultiValueField):
    """
    Some size input fields must be specified per Bootstrap breakpoint. Use this multiple
    input field to handle this.
    """
    def __init__(self, properties, sublabels=None, *args, **kwargs):
        required = kwargs.pop('required', False)
        require_all_fields = kwargs.pop('require_all_fields', required)
        initial = kwargs.pop('initial', None)
        if isinstance(initial, (list, tuple)):
            if len(initial) != len(properties):
                raise ValueError("The number of initial values must be {}.".format(len(properties)))
            initial = dict(zip(properties, initial))
        elif not isinstance(initial, dict):
            initial = {prop: initial for prop in properties}
        allowed_units = kwargs.pop('allowed_units', None)
        fields = [SizeField(required=required, allowed_units=allowed_units)] * len(properties)
        if sublabels is None:
            sublabels = properties
        widget = MultipleTextInputWidget(sublabels)
        super().__init__(fields=fields, widget=widget, required=required,
                         require_all_fields=require_all_fields, initial=initial, *args, **kwargs)
        self.properties = list(properties)

    def prepare_value(self, value):
        """Transform dict from DB into list"""
        if isinstance(value, dict):
            return [value.get(prop) for prop in self.properties]
        return value

    def compress(self, data_list):
        """Transform list into dict for DB"""
        return {prop: value for prop, value in zip(self.properties, data_list)}


class HiddenDictField(Field):
    widget = widgets.HiddenInput

    def prepare_value(self, value):
        """Transform dict from DB into list"""
        if isinstance(value, dict):
            return json.dumps(value)
        return value

    def clean(self, value):
        try:
            return json.loads(value)
        except:
            raise ValidationError("Invalid Field Value")


class CascadeImageField(AdminImageFormField):
    def __init__(self, *args, **kwargs):
        model_name_tuple = filer_settings.FILER_IMAGE_MODEL.split('.')
        Image = apps.get_model(*model_name_tuple, require_ready=False)
        kwargs.setdefault('label', _("Image"))
        super().__init__(
            ManyToOneRel(FilerImageField, Image, 'file_ptr'),
            Image.objects.all(),
            'image_file', *args, **kwargs)



# Generated by Django 3.1.5 on 2021-01-28 15:52

from django.db import migrations, models


def backwards(apps, schema_editor):
    print("Migration backward will not restore your `JSONField`s to `CharField`s.")


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0028_cascade_clipboard'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cascadeelement',
            name='glossary',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='cascadeclipboard',
            name='data',
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name='cascadepage',
            name='glossary',
            field=models.JSONField(blank=True, default=dict, help_text='Store for arbitrary page data.'),
        ),
        migrations.AlterField(
            model_name='cascadepage',
            name='settings',
            field=models.JSONField(blank=True, default=dict, help_text='User editable settings for this page.'),
        ),
        migrations.AlterField(
            model_name='iconfont',
            name='config_data',
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name='inlinecascadeelement',
            name='glossary',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='pluginextrafields',
            name='css_classes',
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name='pluginextrafields',
            name='inline_styles',
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name='sharedglossary',
            name='glossary',
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name='sortableinlinecascadeelement',
            name='glossary',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.RunPython(migrations.RunPython.noop, reverse_code=backwards),
    ]



from django.forms import widgets, BooleanField, CharField
from django.forms.fields import IntegerField
from django.utils.translation import ngettext_lazy, gettext_lazy as _
from django.utils.safestring import mark_safe
from django.utils.text import Truncator
from django.utils.html import escape
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.forms import ManageChildrenFormMixin
from cmsplugin_cascade.plugin_base import TransparentWrapper, TransparentContainer
from cmsplugin_cascade.widgets import NumberInputWidget
from .plugin_base import BootstrapPluginBase


class AccordionFormMixin(ManageChildrenFormMixin, EntangledModelFormMixin):
    num_children = IntegerField(
        min_value=1,
        initial=1,
        widget=NumberInputWidget(attrs={'size': '3', 'style': 'width: 5em !important;'}),
        label=_("Groups"),
        help_text=_("Number of groups for this accordion."),
    )

    close_others = BooleanField(
         label=_("Close others"),
         initial=True,
         required=False,
         help_text=_("Open only one card at a time.")
    )

    first_is_open = BooleanField(
         label=_("First open"),
         initial=True,
         required=False,
         help_text=_("Start with the first card open.")
    )

    class Meta:
        untangled_fields = ['num_children']
        entangled_fields = {'glossary': ['close_others', 'first_is_open']}


class BootstrapAccordionPlugin(TransparentWrapper, BootstrapPluginBase):
    name = _("Accordion")
    default_css_class = 'accordion'
    require_parent = True
    parent_classes = ['BootstrapColumnPlugin']
    direct_child_classes = ['BootstrapAccordionGroupPlugin']
    allow_children = True
    form = AccordionFormMixin
    render_template = 'cascade/bootstrap4/{}accordion.html'

    @classmethod
    def get_identifier(cls, obj):
        num_cards = obj.get_num_children()
        content = ngettext_lazy('with {0} card', 'with {0} cards', num_cards).format(num_cards)
        return mark_safe(content)

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapAccordionPlugin, self).render(context, instance, placeholder)
        context.update({
            'close_others': instance.glossary.get('close_others', True),
            'first_is_open': instance.glossary.get('first_is_open', True),
        })
        return context

    def save_model(self, request, obj, form, change):
        wanted_children = int(form.cleaned_data.get('num_children'))
        super().save_model(request, obj, form, change)
        self.extend_children(obj, wanted_children, BootstrapAccordionGroupPlugin)

plugin_pool.register_plugin(BootstrapAccordionPlugin)


class AccordionGroupFormMixin(EntangledModelFormMixin):
    heading = CharField(
        label=_("Heading"),
        widget=widgets.TextInput(attrs={'size': 80}),
    )

    body_padding = BooleanField(
         label=_("Body with padding"),
         initial=True,
         required=False,
         help_text=_("Add standard padding to card body."),
    )

    class Meta:
        entangled_fields = {'glossary': ['heading', 'body_padding']}

    def clean_heading(self):
        return escape(self.cleaned_data['heading'])


class BootstrapAccordionGroupPlugin(TransparentContainer, BootstrapPluginBase):
    name = _("Accordion Group")
    direct_parent_classes = parent_classes = ['BootstrapAccordionPlugin']
    render_template = 'cascade/generic/naked.html'
    require_parent = True
    form = AccordionGroupFormMixin
    alien_child_classes = True

    @classmethod
    def get_identifier(cls, instance):
        heading = instance.glossary.get('heading', '')
        return Truncator(heading).words(3, truncate=' ...')

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapAccordionGroupPlugin, self).render(context, instance, placeholder)
        context.update({
            'heading': mark_safe(instance.glossary.get('heading', '')),
            'no_body_padding': not instance.glossary.get('body_padding', True),
        })
        return context

plugin_pool.register_plugin(BootstrapAccordionGroupPlugin)



from django.core.exceptions import ValidationError
from django.forms import widgets
from django.forms.fields import CharField, BooleanField, MultiValueField
from django.utils.translation import gettext_lazy as _


class SharedSettingsWidget(widgets.MultiWidget):
    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/sharedsettingsfield.js']

    def __init__(self):
        widget_list = [
            widgets.CheckboxInput(),
            widgets.TextInput(),
        ]
        super().__init__(widget_list)

    def decompress(self, value):
        return value


class SharedSettingsField(MultiValueField):
    def __init__(self, *args, **kwargs):
        kwargs.pop('required', None)
        fields = [
            BooleanField(required=False),
            CharField(required=False),
        ]
        widget = kwargs.pop('widget', SharedSettingsWidget)
        initial = [False, '']
        super().__init__(fields=fields, widget=widget, initial=initial, required=False, *args, **kwargs)

    def clean(self, value):
        if value[0] and not value[1]:
            msg = _("An identifier is required to remember these settings.")
            raise ValidationError(msg)
        return value[1]



from django.utils.safestring import mark_safe

CASCADE_PLUGINS = ['map']

def set_defaults(config):
    config.setdefault('leaflet', {})
    config['leaflet'].setdefault('tilesURL', 'https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png'),
    config['leaflet'].setdefault('default_position', {'lat': 30.0, 'lng': -40.0, 'zoom': 3})
    config['leaflet'].setdefault('id', 'mapbox/streets-v11'),
    config['leaflet'].setdefault('maxZoom', 18),
    config['leaflet'].setdefault('tileSize', 512)
    config['leaflet'].setdefault('zoomOffset', -1)
    config['leaflet'].setdefault('detectRetina', True)
    config['leaflet'].setdefault('attribution', mark_safe('Map data &copy; <a href="http://openstreetmap.org">OpenStreetMap</a>')),
    config['leaflet'].setdefault('addressLookupURL', 'https://nominatim.openstreetmap.org/search'),



from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin
from cmsplugin_cascade.icon.forms import IconFormMixin
from cmsplugin_cascade.icon.plugin_base import IconPluginMixin


class TextIconPlugin(IconPluginMixin, LinkPluginBase):
    """
    This plugin is intended to be used inside the django-CMS-CKEditor.
    """
    name = _("Icon in text")
    text_enabled = True
    render_template = 'cascade/plugins/texticon.html'
    ring_plugin = 'IconPlugin'
    parent_classes = ['TextPlugin']
    form = type('TextIconForm', (LinkFormMixin, IconFormMixin), {'require_link': False})
    model_mixins = (LinkElementMixin,)
    allow_children = False
    require_parent = False

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/iconplugin.js']

    @classmethod
    def requires_parent_plugin(cls, slot, page):
        return False

plugin_pool.register_plugin(TextIconPlugin)



from django.core.exceptions import ValidationError
from django.db.models import Q
from django.forms import widgets
from django.forms.fields import BooleanField, ChoiceField, MultipleChoiceField
from django.utils.safestring import mark_safe
from django.utils.text import format_lazy
from django.utils.translation import ngettext_lazy, gettext_lazy as _

from cms.plugin_pool import plugin_pool
from entangled.forms import EntangledModelFormMixin
from cmsplugin_cascade import app_settings
from cmsplugin_cascade.bootstrap4.grid import Breakpoint
from cmsplugin_cascade.forms import ManageChildrenFormMixin
from .plugin_base import BootstrapPluginBase
from . import grid


def get_widget_choices():
    breakpoints = app_settings.CMSPLUGIN_CASCADE['bootstrap4']['fluid_bounds']
    widget_choices = []
    for index, (bp, bound) in enumerate(breakpoints.items()):
        if index == 0:
            widget_choices.append((bp.name, "{} (<{:.1f}px)".format(bp.label, bound.max)))
        elif index == len(breakpoints) - 1:
            widget_choices.append((bp.name, "{} (≥{:.1f}px)".format(bp.label, bound.min)))
        else:
            widget_choices.append((bp.name, "{} (≥{:.1f}px and <{:.1f}px)".format(bp.label, bound.min, bound.max)))
    return widget_choices


class ContainerBreakpointsWidget(widgets.CheckboxSelectMultiple):
    template_name = 'cascade/admin/widgets/container_breakpoints.html'


class ContainerFormMixin(EntangledModelFormMixin):
    breakpoints = MultipleChoiceField(
        label=_('Available Breakpoints'),
        choices=get_widget_choices(),
        widget=ContainerBreakpointsWidget(choices=get_widget_choices()),
        initial=[bp.name for bp in app_settings.CMSPLUGIN_CASCADE['bootstrap4']['fluid_bounds'].keys()],
        help_text=_("Supported display widths for Bootstrap's grid system."),
    )

    fluid = BooleanField(
        label=_('Fluid Container'),
        initial=False,
        required=False,
        help_text=_("Changing your outermost '.container' to '.container-fluid'.")
    )

    class Meta:
        entangled_fields = {'glossary': ['breakpoints', 'fluid']}

    def clean_breapoints(self):
        # TODO: check this
        if len(self.cleaned_data['glossary']['breakpoints']) == 0:
            raise ValidationError(_("At least one breakpoint must be selected."))
        return self.cleaned_data['glossary']


class ContainerGridMixin:
    def get_grid_instance(self):
        fluid = self.glossary.get('fluid', False)
        try:
            breakpoints = [getattr(grid.Breakpoint, bp) for bp in self.glossary['breakpoints']]
        except KeyError:
            breakpoints = [bp for bp in grid.Breakpoint]
        if fluid:
            bounds = dict((bp, grid.fluid_bounds[bp]) for bp in breakpoints)
        else:
            bounds = dict((bp, grid.default_bounds[bp]) for bp in breakpoints)
        return grid.Bootstrap4Container(bounds=bounds)


class BootstrapContainerPlugin(BootstrapPluginBase):
    name = _("Container")
    parent_classes = None
    require_parent = False
    model_mixins = (ContainerGridMixin,)
    form = ContainerFormMixin
    footnote_html = """<p>
    For more information about the Container please read the
    <a href="https://getbootstrap.com/docs/4.3/layout/overview/#containers" target="_new">Bootstrap documentation</a>.
    </p>"""

    @classmethod
    def get_identifier(cls, obj):
        breakpoints = obj.glossary.get('breakpoints')
        content = obj.glossary.get('fluid') and '(fluid) ' or ''
        if breakpoints:
            BREAKPOINTS = app_settings.CMSPLUGIN_CASCADE['bootstrap4']['fluid_bounds']
            devices = ', '.join([str(bp.label) for bp in BREAKPOINTS if bp.name in breakpoints])
            content = _("{0}for {1}").format(content, devices)
        return mark_safe(content)

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapContainerPlugin, cls).get_css_classes(obj)
        if obj.glossary.get('fluid'):
            css_classes.append('container-fluid')
        else:
            css_classes.append('container')
        return css_classes

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        obj.sanitize_children()

plugin_pool.register_plugin(BootstrapContainerPlugin)


class BootstrapRowFormMixin(ManageChildrenFormMixin, EntangledModelFormMixin):
    """
    Form class to add non-materialized field to count the number of children.
    """
    ROW_NUM_COLUMNS = [1, 2, 3, 4, 6, 12]
    num_children = ChoiceField(
        label=_('Columns'),
        choices=[(i, ngettext_lazy('{0} column', '{0} columns', i).format(i)) for i in ROW_NUM_COLUMNS],
        initial=3,
        help_text=_('Number of columns to be created with this row.'),
    )

    class Meta:
        untangled_fields = ['num_children']


class RowGridMixin:
    def get_grid_instance(self):
        row = grid.Bootstrap4Row()
        query = Q(plugin_type='BootstrapContainerPlugin') | Q(plugin_type='BootstrapColumnPlugin') \
          | Q(plugin_type='BootstrapJumbotronPlugin')
        container = self.get_ancestors().order_by('depth').filter(query).last().get_bound_plugin().get_grid_instance()
        container.add_row(row)
        return row


class BootstrapRowPlugin(BootstrapPluginBase):
    name = _("Row")
    default_css_class = 'row'
    parent_classes = ['BootstrapContainerPlugin', 'BootstrapColumnPlugin', 'BootstrapJumbotronPlugin']
    model_mixins = (RowGridMixin,)
    form = BootstrapRowFormMixin

    @classmethod
    def get_identifier(cls, obj):
        num_cols = obj.get_num_children()
        content = ngettext_lazy("with {0} column", "with {0} columns", num_cols).format(num_cols)
        return mark_safe(content)

    def save_model(self, request, obj, form, change):
        wanted_children = int(form.cleaned_data.get('num_children'))
        super().save_model(request, obj, form, change)
        child_glossary = {'xs-column-width': 'col'}
        self.extend_children(obj, wanted_children, BootstrapColumnPlugin, child_glossary=child_glossary)

plugin_pool.register_plugin(BootstrapRowPlugin)


class ColumnGridMixin:
    valid_keys = ['xs-column-width', 'sm-column-width', 'md-column-width', 'lg-column-width', 'xs-column-width',
                  'xs-column-offset', 'sm-column-offset', 'md-column-offset', 'lg-column-offset', 'xs-column-offset']
    def get_grid_instance(self):
        column = None
        query = Q(plugin_type='BootstrapRowPlugin')
        row_obj = self.get_ancestors().order_by('depth').filter(query).last().get_bound_plugin()
        # column_siblings = row_obj.get_descendants().order_by('depth').filter(plugin_type='BootstrapColumnPlugin')
        row = row_obj.get_grid_instance()
        for column_sibling in self.get_siblings():
            classes = [val for key, val in column_sibling.get_bound_plugin().glossary.items()
                       if key in self.valid_keys and val]
            if column_sibling.pk == self.pk:
                column = grid.Bootstrap4Column(classes)
                row.add_column(column)
            else:
                row.add_column(grid.Bootstrap4Column(classes))
        return column


class BootstrapColumnPlugin(BootstrapPluginBase):
    name = _("Column")
    parent_classes = ['BootstrapRowPlugin']
    child_classes = ['BootstrapJumbotronPlugin']
    alien_child_classes = True
    default_css_attributes = [fmt.format(bp.name) for bp in grid.Breakpoint
        for fmt in ('{}-column-width', '{}-column-offset', '{}-column-ordering', '{}-responsive-utils')]
    model_mixins = (ColumnGridMixin,)

    def get_form(self, request, obj=None, **kwargs):
        def choose_help_text(*phrases):
            bounds = 'fluid_bounds' if container.glossary.get('fluid') else 'default_bounds'
            bs4_breakpoints = app_settings.CMSPLUGIN_CASCADE['bootstrap4'][bounds]
            if last:
                return phrases[0].format(bs4_breakpoints[last].max)
            elif len(breakpoints) > 1:
                return phrases[1].format(bs4_breakpoints[first].min)
            else:
                return phrases[2]

        if 'parent' in self._cms_initial_attributes:
            container=self._cms_initial_attributes['parent'].get_ancestors().order_by('depth').last().get_bound_plugin()
        else:
            containers=obj.get_ancestors().filter(plugin_type='BootstrapContainerPlugin')
            if containers:
                container=containers.order_by('depth').last().get_bound_plugin()
            else:
                jumbotrons=obj.get_ancestors().filter(plugin_type='BootstrapJumbotronPlugin')
                container=jumbotrons.order_by('depth').last().get_bound_plugin()
        breakpoints = container.glossary['breakpoints']

        width_fields, offset_fields, reorder_fields, responsive_fields = {}, {}, {}, {}
        units = [ngettext_lazy("{} unit", "{} units", i).format(i) for i in range(0, 13)]
        previous_devices, previous_label = '', ''
        for bp in breakpoints:
            try:
                last = getattr(grid.Breakpoint, breakpoints[breakpoints.index(bp)])
            except IndexError:
                last = None
            finally:
                first = getattr(grid.Breakpoint, bp)
                devices = ', '.join([str(b.label) for b in grid.Breakpoint.range(first, last)])

            if bp == 'xs':
                choices = [('col', _("Flex column"))]
                choices.extend(('col-{}'.format(i), _("{} fixed column").format(units[i])) for i in range(1, 13))
                choices.append(('col-auto', _("Auto column")))
            else:
                choices = [('col-{}'.format(bp), _("Flex column"))]
                choices.extend(('col-{}-{}'.format(bp, i), _("{} fixed column").format(units[i])) for i in range(1, 13))
                choices.append(('col-{}-auto'.format(bp), _("Auto column")))
            if breakpoints.index(bp) == 0:
                # first breakpoint
                field_name = '{}-column-width'.format(bp)
                width_fields[field_name] = ChoiceField(
                    choices=choices,
                    label=_("Column width for {}").format(devices),
                    initial='col' if bp == 'xs' else 'col-{}'.format(bp),
                    help_text=choose_help_text(
                        _("Column width for devices narrower than {:.1f} pixels."),
                        _("Column width for devices wider than {:.1f} pixels."),
                        _("Column width for all devices."),
                    )
                )
            else:
                # wider breakpoints may inherit from next narrower ones
                choices.insert(0, ('', format_lazy(_("Inherit column width from {}"), previous_devices)))
                field_name = '{}-column-width'.format(bp)
                width_fields[field_name] = ChoiceField(
                    choices=choices,
                    label=_("Column width for {}").format(devices),
                    initial='',
                    required=False,
                    help_text=choose_help_text(
                        _("Override column width for devices narrower than {:.1f} pixels."),
                        _("Override column width for devices wider than {:.1f} pixels."),
                        _("Override column width for all devices."),
                    )
                )
            previous_devices = devices

            # handle offset
            if breakpoints.index(bp) == 0:
                choices = [('', _("No offset"))]
                offset_range = range(1, 13)
            else:
                choices = [('', format_lazy(_("Inherit offset from {}"), previous_label))]
                offset_range = range(0, 13)
            previous_label = Breakpoint[bp].label
            if bp == 'xs':
                choices.extend(('offset-{}'.format(i), units[i]) for i in offset_range)
            else:
                choices.extend(('offset-{}-{}'.format(bp, i), units[i]) for i in offset_range)
            label = _("Offset for {}").format(devices)
            help_text = choose_help_text(
                _("Offset width for devices narrower than {:.1f} pixels."),
                _("Offset width for devices wider than {:.1f} pixels."),
                _("Offset width for all devices.")
            )
            field_name = '{}-column-offset'.format(bp)
            offset_fields[field_name] = ChoiceField(
                choices=choices,
                label=label,
                required=False,
                help_text=help_text,
            )

            # handle column reordering
            choices = [('', _("No reordering"))]
            if bp == 'xs':
                choices.extend(('order-{}'.format(i), _("Reorder by {}").format(units[i])) for i in range(1, 13))
            else:
                choices.extend(('order-{}-{}'.format(bp, i), _("Reorder by {}").format(units[i])) for i in range(1, 13))
            label = _("Reordering for {}").format(devices)
            help_text = choose_help_text(
                _("Reordering for devices narrower than {:.1f} pixels."),
                _("Reordering for devices wider than {:.1f} pixels."),
                _("Reordering for all devices.")
            )
            field_name = '{}-column-ordering'.format(bp)
            reorder_fields[field_name] = ChoiceField(
                choices=choices,
                label=label,
                required=False,
                help_text=help_text,
            )

            # handle responsive utilities
            choices = [('', _("Default")), ('visible-{}'.format(bp), _("Visible")), ('hidden-{}'.format(bp), _("Hidden"))]
            label = _("Responsive utilities for {}").format(devices)
            help_text = choose_help_text(
                _("Utility classes for showing and hiding content by devices narrower than {:.1f} pixels."),
                _("Utility classes for showing and hiding content by devices wider than {:.1f} pixels."),
                _("Utility classes for showing and hiding content for all devices.")
            )
            field_name = '{}-responsive-utils'.format(bp)
            responsive_fields[field_name] = ChoiceField(
                choices=choices,
                label=label,
                initial='',
                widget=widgets.RadioSelect,
                required=False,
                help_text=help_text,
            )
        glossary_fields = list(width_fields.keys())
        glossary_fields.extend(offset_fields.keys())
        glossary_fields.extend(reorder_fields.keys())
        glossary_fields.extend(responsive_fields.keys())

        class Meta:
            entangled_fields = {'glossary': glossary_fields}

        attrs = dict(width_fields, **offset_fields, **reorder_fields, **responsive_fields, Meta=Meta)
        kwargs['form'] = type('ColumnForm', (EntangledModelFormMixin,), attrs)
        return super().get_form(request, obj, **kwargs)

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        obj.sanitize_children()

    @classmethod
    def sanitize_model(cls, obj):
        sanitized = super().sanitize_model(obj)
        return sanitized

    @classmethod
    def get_identifier(cls, obj):
        glossary = obj.get_complete_glossary()
        widths = []
        for bp in glossary.get('breakpoints', []):
            width = obj.glossary.get('{0}-column-width'.format(bp), '').replace('col-{0}-'.format(bp), '')
            if width:
                widths.append(width)
        if len(widths) > 0:
            content = _("widths: {}").format(' / '.join(widths))
        else:
            content = _("unknown width")
        return mark_safe(content)

plugin_pool.register_plugin(BootstrapColumnPlugin)
